# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2013, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
import os
import tempfile
import shutil
import unittest
import zipfile


here = os.path.abspath(os.path.dirname(__file__))
TEST_DATA_DIR = os.path.join(here, 'tests', 'data')


class EPUBTestCase(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tmpdir)

    def pack_epub(self, directory):
        """Given an directory containing epub contents,
        pack it up and make return filepath.
        Packed file is remove on test exit.
        """
        zip_fd, zip_filepath = tempfile.mkstemp('.epub', dir=self.tmpdir)
        with zipfile.ZipFile(zip_filepath, 'w') as zippy:
            base_path = os.path.abspath(directory)
            for root, dirs, filenames in os.walk(directory):
                # Strip the absolute path
                archive_path = os.path.abspath(root)[len(base_path):]
                for filename in filenames:
                    filepath = os.path.join(root, filename)
                    archival_filepath = os.path.join(archive_path, filename)
                    zippy.write(filepath, archival_filepath)
        return zip_filepath

    def copy(self, src, dst_name='book'):
        """Convenient method for copying test data directories."""
        dst = os.path.join(self.tmpdir, dst_name)
        shutil.copytree(src, dst)
        return dst



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2013, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
import os
import io
import tempfile
import zipfile
from collections import Sequence

import jinja2
from lxml import etree


__all__ = ('EPUB', 'Package', 'Item',)


# ./mimetype
EPUB_MIMETYPE_RELATIVE_PATH = "mimetype"
EPUB_MIMETYPE_CONTENTS = "application/epub+zip"
# ./META-INF/container.xml
EPUB_CONTAINER_XML_RELATIVE_PATH = "META-INF/container.xml"
EPUB_CONTAINER_XML_NAMESPACES = {
    'ns': "urn:oasis:names:tc:opendocument:xmlns:container"
    }
# ./*.opf
EPUB_OPF_NAMESPACES = {
    'opf': "http://www.idpf.org/2007/opf",
    'dc': "http://purl.org/dc/elements/1.1/",
    'lrmi': "http://lrmi.net/the-specification",
    }
OPF_TEMPLATE = """\
{% set metadata = package.metadata %}
<?xml version="1.0" encoding="UTF-8"?>
<package xmlns="http://www.idpf.org/2007/opf"
         xml:lang="en"
         version="3.0"
         unique-identifier="pub-id"
         prefix="cc: http://creativecommons.org/ns#"
         >
  <metadata xmlns:dc="http://purl.org/dc/elements/1.1/">
    <dc:title id="title">{{ metadata['title'] }}</dc:title>
    <meta property="title-type"
          refines="#title"
          >main</meta>
    <dc:creator id="creator"
                file-as="{{ metadata['publisher'] }}"
                >{{ metadata['publisher'] }}</dc:creator>
    <dc:identifier id="pub-id">{{ metadata['identifier'] }}</dc:identifier>
    <dc:language>{{ metadata['language'] }}</dc:language>
    <dc:publisher>{{ metadata['publisher'] }}</dc:publisher>
    <meta property="publicationMessage">{{ metadata['publication_message'] }}</meta>
    {% if 'license_text' in metadata %}
    <dc:rights>{{ metadata['license_text'] }}</dc:rights>
    {% endif %}
    <link rel="cc:license" href="{{ metadata['license_url'] }}"/>
    <meta property="cc:attributionURL">http://cnx.org/contents</meta>
  </metadata>
  <manifest>
    {% for item in package -%}
      <item media-type="{{ item.media_type }}"
            href="{{ locations[item] }}"
            {% if item.is_navigation %}
            id="toc"
            {% endif %}
            {% if item.properties is defined and item.properties %}
            properties="{{ ', '.join(item.properties) }}"
            {% endif %}
            />
    {%- endfor %}
  </manifest>
</package>
"""
CONTAINER_XML_TEMPLATE = """\
<?xml version="1.0" encoding="UTF-8"?>
<container xmlns="urn:oasis:names:tc:opendocument:xmlns:container"
           version="1.0">
  <rootfiles>
    {% for filename in package_filenames -%}
    <rootfile media-type="application/oebps-package+xml"
              full-path="{{ filename }}"
              />
    {%- endfor %}
  </rootfiles>
</container>
"""


class MissingNavigationError(Exception):
    """Raised when a ``Package`` is missing a navigation document.
    http://www.idpf.org/epub/30/spec/epub30-overview.html#sec-nav
    """


class AdditionalNavigationError(Exception):
    """Raised when a ``Package`` has more then one navigation document."""


class MissingMetadataError(Exception):
    """Raised when a piece of required metadata is missing from the document.
    """


def pack_epub(directory, file):
    """Pack the given ``directory`` into an epub (i.e. zip) archive
    given as ``file``, which can be a file-path or file-like object.
    """
    with zipfile.ZipFile(file, 'w') as zippy:
        base_path = os.path.abspath(directory)
        for root, dirs, filenames in os.walk(directory):
            # Strip the absolute path
            archive_path = os.path.relpath(root, base_path)
            for filename in filenames:
                filepath = os.path.join(root, filename)
                archival_filepath = os.path.join(archive_path, filename)
                zippy.write(filepath, archival_filepath)


def unpack_epub(file, directory):
    """Unpack the given ``file`` (a file-path or file-like object)
    to the given ``directory``.
    """
    if zipfile.is_zipfile(file):
        # Extract the epub to the current working directory.
        with zipfile.ZipFile(file, 'r') as zf:
            zf.extractall(path=directory)


class EPUB(Sequence):
    """Represents an EPUB3 file structure in object form.
    It is designed to work with .epub files (zip files),
    but will uncompress them to a temporary location before working with them.

    """

    def __init__(self, packages=None, root=None):
        self._packages = packages is None and [] or packages
        self._root = root

    @classmethod
    def from_file(cls, file):
        """Create the object from a *file* or *file-like object*.
        The file can point to an ``.epub`` file or a directory
        (the contents of which reflect
        the internal struture of an ``.epub`` archive).
        If given an non-archive file,
        this structure will be used when reading in and parsing the epub.
        If an archive file is given,
        it will be extracted to the temporal filesystem.
        """
        root = None
        if zipfile.is_zipfile(file):
            unpack_dir = tempfile.mkdtemp('-epub')
            # Extract the epub to the current working directory.
            with zipfile.ZipFile(file, 'r') as zf:
                zf.extractall(path=unpack_dir)
            root = unpack_dir
        elif os.path.isdir(file):
            root = file
        else:
            raise TypeError("Can't decipher what should be done "
                            "with the given file.")

        # NOTE We ignore the mimetype file, as it's not extremely important
        #      to anything done here.

        # Build a blank epub object then parse the packages.
        container_xml_filepath = os.path.join(root,
                                              EPUB_CONTAINER_XML_RELATIVE_PATH)
        container_xml = etree.parse(container_xml_filepath)

        packages = []
        for pkg_filepath in container_xml.xpath(
                '//ns:rootfile/@full-path',
                namespaces=EPUB_CONTAINER_XML_NAMESPACES):
            filepath = os.path.join(root, pkg_filepath)
            packages.append(Package.from_file(filepath))
        return cls(packages=packages, root=root)

    @staticmethod
    def to_file(epub, file):
        """Export to ``file``, which is a *file* or *file-like object*."""
        directory = tempfile.mkdtemp('-epub')
        # Write out the contents to the filesystem.
        package_filenames = []
        for package in epub:
            opf_filepath = Package.to_file(package, directory)
            opf_filename = os.path.basename(opf_filepath)
            package_filenames.append(opf_filename)

        # Create the container.xml
        container_xml_filepath = os.path.join(directory,
                                              EPUB_CONTAINER_XML_RELATIVE_PATH)
        template = jinja2.Template(CONTAINER_XML_TEMPLATE,
                                   trim_blocks=True, lstrip_blocks=True)
        os.makedirs(os.path.dirname(container_xml_filepath))  # FIXME PY3
        with open(container_xml_filepath, 'w') as fb:
            xml  = template.render(package_filenames=package_filenames)
            fb.write(xml)
        # Write the mimetype file.
        with open(os.path.join(directory, 'mimetype'), 'w') as fb:
            fb.write("application/epub+zip")

        # Pack everything up
        pack_epub(directory, file=file)

    # ABC methods for MutableSequence
    def __getitem__(self, k):
        return self._packages[k]

    def __len__(self):
        return len(self._packages)


class OPFParser:
    """Parse an ``.opf`` xml document's metadata.
    This class is callable to respond as a function
    after inititiation as a singleton.
    The work for parsing the metadata is methodized to provide
    detailed exceptions.

    """
    namespaces = EPUB_OPF_NAMESPACES
    metadata_required_keys = (
        'publisher', 'publication_message',
        )
    metadata_optional_keys = (
        'title', 'identifier', 'language', 'license_text', 'license_url',
        )

    def __init__(self, opf_xml):
        """Given the opf_xml (an ``lxml.etree.ElementTree``),
        parse the metadata fields on access.
        """
        self._xml = opf_xml

    def parser(self, xpath, prefix="/opf:package/opf:metadata/"):
        metadata_prefix = "/opf:package/opf:metadata/"
        values = self._xml.xpath(metadata_prefix + xpath,
                                 namespaces=self.namespaces)
        return values

    @property
    def title(self):
        items = self.parser('dc:title/text()')
        try:
            value = items[0]
        except IndexError:
            value = None
        return value

    @property
    def publisher(self):
        items = self.parser('dc:creator/text()')
        try:
            value = items[0]
        except IndexError:
            value = None
        return value

    @property
    def publication_message(self):
        items = self.parser('opf:meta[@property="publicationMessage"]/text()')
        try:
            value = items[0]
        except IndexError:
            value = None
        return value

    @property
    def identifier(self):
        items = self.parser('dc:identifier/text()')
        try:
            value = items[0]
        except IndexError:
            value = None
        return value

    @property
    def language(self):
        items = self.parser('dc:language/text()')
        try:
            value = items[0]
        except IndexError:
            value = None
        return value

    @property
    def license_text(self):
        items = self.parser('dc:rights/text()')
        try:
            value = items[0]
        except IndexError:
            value = None
        return value

    @property
    def license_url(self):
        items = self.parser('opf:link[@rel="cc:license"]/@href')
        try:
            value = items[0]
        except IndexError:
            value = None
        return value

    @property
    def metadata(self):
        items = {}
        keyrings = (self.metadata_required_keys, self.metadata_optional_keys,)
        for keyring in keyrings:
            for key in keyring:
                # TODO On refactoring of the metadata properties,
                # required fields will raise MissingMetadataError,
                # but for now do that here.
                value = getattr(self, key)
                if key in self.metadata_required_keys and value is None:
                    raise MissingMetadataError(
                        "A value for '{}' could not be found.".format(key))
                elif value is None:
                    continue
                items[key] = value
        return items


class Package(Sequence):
    """EPUB3 package"""

    def __init__(self, name, items, metadata=None):
        self.name = name
        self.metadata = metadata or {}
        self._items = items
        navigation_items = [i for i in self._items if i.is_navigation]
        if len(navigation_items) == 0:
            raise MissingNavigationError("Navigation item not found")
        elif len(navigation_items) > 1:
            raise AdditionalNavigationError(
                "Only one navigation item can exist "
                "per package. The given value is a second "
                "navigation item.")
        else:
            index = self._items.index(navigation_items[0])
            self._navigation_item_index = index

    @classmethod
    def from_file(cls, file):
        """Create the object from a *file* or *file-like object*."""
        opf_xml = etree.parse(file)
        # Check if ``file`` is file-like.
        if hasattr(file, 'read'):
            name = os.path.basename(file.name)
            root = os.path.abspath(os.path.dirname(file.name))
        else:  # ...a filepath
            name = os.path.basename(file)
            root = os.path.abspath(os.path.dirname(file))
        parser = OPFParser(opf_xml)

        # Roll through the item entries
        manifest = opf_xml.xpath('/opf:package/opf:manifest/opf:item',
                                 namespaces=EPUB_OPF_NAMESPACES)
        pkg_items = []
        for item in manifest:
            absolute_filepath = os.path.join(root, item.get('href'))
            properties = item.get('properties', '').split()
            is_navigation = 'nav' in properties
            media_type = item.get('media-type')
            pkg_items.append(Item.from_file(absolute_filepath,
                                          media_type=media_type,
                                          is_navigation=is_navigation,
                                          properties=properties))
        # Ignore spine ordering, because it is not important
        #   for our use cases.
        return cls(name, pkg_items, parser.metadata)

    @staticmethod
    def to_file(package, directory):
        """Write the package to the given ``directory``.
        Returns the OPF filename.
        """
        opf_filepath = os.path.join(directory, package.name)

        # Create the directory structure
        for name in ('contents', 'resources',):
            path = os.path.join(directory, name)
            if not os.path.exists(path):
                os.mkdir(path)

        # Write the items to the filesystem
        locations = {}  # Used when rendering
        for item in package:
            if item.media_type == 'application/xhtml+xml':
                base = os.path.join(directory, 'contents')
            else:
                base = os.path.join(directory, 'resources')
            filename = item.name
            filepath = os.path.join(base, filename)
            locations[item] = os.path.relpath(filepath, directory)
            with open(filepath, 'wb') as item_file:
                item_file.write(item.data.read())

        # Write the OPF
        template = jinja2.Template(OPF_TEMPLATE,
                                   trim_blocks=True, lstrip_blocks=True)
        with open(opf_filepath, 'wb') as opf_file:
            opf = template.render(package=package, locations=locations)
            if not isinstance(opf, bytes):
                opf = opf.encode('utf-8')
            opf_file.write(opf)

        return opf_filepath

    @property
    def navigation(self):
        return self._items[self._navigation_item_index]

    def grab_by_name(self, name):
        try:
            return [i for i in self._items if i.name == name][0]
        except IndexError:
            raise KeyError("'{}' not found in package.".format(name))

    # ABC methods for Sequence
    def __getitem__(self, k):
        return self._items[k]

    def __len__(self):
        return len(self._items)


class Item:
    """Package item"""

    def __init__(self, name, data=None, media_type=None,
                 is_navigation=False, properties=[], **kwargs):
        self.name = name
        self.data = data
        self.media_type = media_type
        self.is_navigation = bool(is_navigation)
        self.properties = properties

    @classmethod
    def from_file(cls, filepath, **kwargs):
        name = os.path.basename(filepath)
        with open(filepath, 'rb') as fb:
            data = io.BytesIO(fb.read())
        return cls(name, data, **kwargs)



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2013, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
import io
import hashlib
import mimetypes
try:
    from collections.abc import MutableSequence, Iterable
except ImportError:
    from collections import MutableSequence, Iterable
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse
from contextlib import contextmanager

import lxml.html
import lxml.html.builder
from lxml import etree


__all__ = (
    'TRANSLUCENT_BINDER_ID', 'RESOURCE_HASH_TYPE',
    'INTERNAL_REFERENCE_TYPE', 'EXTERNAL_REFERENCE_TYPE',
    'REFERENCE_REMOTE_TYPES',
    'ATTRIBUTED_ROLE_KEYS',
    'flatten_tree_to_ident_hashes', 'model_to_tree',
    'flatten_model', 'flatten_to_documents',
    'Binder', 'TranslucentBinder', 'Document', 'DocumentPointer', 'Resource',
    )


mimetypes.init()
RESOURCE_HASH_TYPE = 'sha1'
TRANSLUCENT_BINDER_ID = 'subcol'
INTERNAL_REFERENCE_TYPE = 'internal'
EXTERNAL_REFERENCE_TYPE = 'external'
INLINE_REFERENCE_TYPE = 'inline'
REFERENCE_REMOTE_TYPES = (INTERNAL_REFERENCE_TYPE, EXTERNAL_REFERENCE_TYPE, INLINE_REFERENCE_TYPE,)
ATTRIBUTED_ROLE_KEYS = (
    # MUST be alphabetical
    'authors', 'copyright_holders', 'editors', 'illustrators',
    'publishers', 'translators',
    )


def utf8(item):
    if isinstance(item, list):
        return [utf8(i) for i in item]
    if isinstance(item, dict):
        return {utf8(k): utf8(v) for k, v in item.items()}
    try:
        return item.decode('utf-8')
    except:
        return item


def model_to_tree(model, title=None, lucent_id=TRANSLUCENT_BINDER_ID):
    """Given an model, build the tree::

        {'id': <id>|'subcol', 'title': <title>, 'contents': [<tree>, ...]}

    """
    if type(model) is TranslucentBinder:
        id = lucent_id
    else:
        id = model.ident_hash
    title = title is not None and title or model.metadata.get('title')
    tree = {'id': id, 'title': title}
    if hasattr(model, '__iter__'):
        contents = tree['contents'] = []
        for node in model:
            item = model_to_tree(node, model.get_title_for_node(node),
                                 lucent_id=lucent_id)
            contents.append(item)
    return tree


def flatten_tree_to_ident_hashes(item_or_tree, lucent_id=TRANSLUCENT_BINDER_ID):
    """Flatten a tree to id and version values (ident_hash)."""
    if 'contents' in item_or_tree:
        tree = item_or_tree
        if tree['id'] != lucent_id:
            yield tree['id']
        for i in tree['contents']:
            ##yield from flatten_tree_to_ident_hashs(i, lucent_id)
            for x in flatten_tree_to_ident_hashes(i, lucent_id):
                yield x
    else:
        item = item_or_tree
        yield item['id']
    raise StopIteration()


def flatten_model(model):
    """Flatten a model to a list of models.
    This is used to flatten a ``Binder``'ish model down to a list
    of contained models.
    """
    yield model
    if isinstance(model, (TranslucentBinder, Binder,)):
        for m in model:
            ##yield from flatten_model(m)
            for x in flatten_model(m):
                yield x
    raise StopIteration()


def flatten_to_documents(model, include_pointers=False):
    """Flatten the model to a list of documents (aka ``Document`` objects).
    This is to flatten a ``Binder``'ish model down to a list of documents.
    If ``include_pointers`` has been set to ``True``, ``DocumentPointers``
    will also be included in the results.
    """
    for m in flatten_model(model):
        if isinstance(m, Document):
            yield m
        elif include_pointers and isinstance(m, DocumentPointer):
            yield m
    raise StopIteration()


def _discover_uri_type(uri):
    """Given a ``uri``, determine if it is internal or external."""
    parsed_uri = urlparse(uri)
    if not parsed_uri.netloc:
        if parsed_uri.scheme == 'data':
            type_ = INLINE_REFERENCE_TYPE
        else:
            type_ = INTERNAL_REFERENCE_TYPE
    else:
        type_ = EXTERNAL_REFERENCE_TYPE
    return type_

def _parse_references(xml):
    """Parse the references to ``Reference`` instances."""
    references = []
    ref_finder = HTMLReferenceFinder(xml)
    for elm, uri_attr in ref_finder:
        type_ = _discover_uri_type(elm.get(uri_attr))
        references.append(Reference(elm, type_, uri_attr))
    return references


class Reference(object):
    """A reference within a ``Document`` model, either internal or external.
    This depends on an xml element tree, to provide binds for uri and name.
    """

    def __init__(self, elm, remote_type, uri_attr):
        self.elm = elm
        try:
            assert remote_type in REFERENCE_REMOTE_TYPES
        except AssertionError:
            raise ValueError("remote_type: '{}' is invalid." \
                             .format(remote_type))
        self.remote_type = remote_type
        self._uri_attr = uri_attr
        self._bound_model = None
        self._uri_template = None

    @property
    def is_bound(self):
        return self._bound_model is not None

    # read-only property, use bind for writing.
    @property
    def bound_model(self):
        return self._bound_model

    def _get_uri(self):
        if self.is_bound:
            # Update the value before returning.
            self._set_uri_from_bound_model()
        return self.elm.get(self._uri_attr)

    def _set_uri(self, value):
        if self.is_bound:
            raise ValueError("URI is bound to an object. Unbind first.")
        self.elm.set(self._uri_attr, value)

    uri = property(_get_uri, _set_uri)

    @property
    def uri_parts(self):
        """Returns a parsed URI"""
        return urlparse(self.uri)

    def _set_uri_from_bound_model(self):
        """Using the bound model, set the uri."""
        value = self._uri_template.format(self._bound_model.id)
        self.elm.set(self._uri_attr, value)

    def bind(self, model, template="{}"):
        """Bind the ``model`` to the reference. This uses the model's
        ``id`` attribute and the given ``template`` to
        dynamically produce a uri when accessed.
        """
        self._bound_model = model
        self._uri_template = template
        self._set_uri_from_bound_model()

    def unbind(self):
        """Unbind the model from the reference."""
        self._bound_model = None
        self._uri_template = None


class HTMLReferenceFinder(object):
    """Find references within an HTML xml element tree."""

    def __init__(self, xml):
        self.xml = xml

    def __iter__(self):
        for elm in self._archors():
            yield elm, 'href'
        for elm, uri_attr in self._media():
            yield elm, uri_attr
        raise StopIteration()

    def apply_xpath(self, xpath):
        return self.xml.xpath(xpath)

    def _archors(self):
        return self.apply_xpath('//a')

    def _media(self):
        media_xpath = {
                '//img[@src]': 'src',
                '//audio[@src]': 'src',
                '//video[@src]': 'src',
                '//object[@data]': 'data',
                '//object/embed[@src]': 'src',
                '//source[@src]': 'src',
                '//span[@data-src]': 'data-src',
                }
        for xpath, attr in media_xpath.items():
            for elm in self.apply_xpath(xpath):
                yield elm, attr
        raise StopIteration()


# ########## #
#   Models   #
# ########## #

class TranslucentBinder(MutableSequence):
    """A clear/translucent binder instance.
    This is used only represent ``Binder`` behavior
    without being a persistent piece of data.
    """
    id = None
    ident_hash = None

    def __init__(self, nodes=None, metadata=None,
                 title_overrides=None):
        self._nodes = nodes or []
        self.metadata = utf8(metadata or {})
        if title_overrides is not None:
            if len(self._nodes) != len(title_overrides):
                raise ValueError(
                    "``title_overrides`` should be the same length as "
                    "``nodes``. {} != {}" \
                    .format(len(self._nodes), len(title_overrides)))
            self._title_overrides = utf8(title_overrides)
        else:
            self._title_overrides = [None] * len(self._nodes)

    @property
    def ident_hash(self):
        return None

    @property
    def is_translucent(self):
        return self.__class__ is TranslucentBinder

    def set_title_for_node(self, node, title):
        index = self._nodes.index(node)
        self._title_overrides[index] = title

    def get_title_for_node(self, node):
        index = self._nodes.index(node)
        return self._title_overrides[index]

    # ABC methods for MutableSequence
    def __getitem__(self, i):
        return self._nodes[i]

    def __setitem__(self, i, v):
        self._nodes[i] = v

    def __delitem__(self, i):
        del self._nodes[i]
        del self._title_overrides[i]

    def __len__(self):
        return len(self._nodes)

    def insert(self, i, v):
        self._nodes.insert(i, v)
        self._title_overrides.insert(i, None)


class Binder(TranslucentBinder):
    """An object that has metadata and contains
    ``Binder``, ``TranslucentBinder`` and ``Document`` instances.
    """

    def __init__(self, id, nodes=None, metadata=None, title_overrides=None):
        super(Binder, self).__init__(nodes, metadata, title_overrides)
        self.id = id

    @property
    def ident_hash(self):
        if self.id is not None:
            args = [self.id]
            version = self.metadata.get('version')
            if version is not None:
                args.append(version)
            value = '@'.join(args)
        else:
            value = None
        return value

    def get_uri(self, system, default=None):
        try:
            uri = self.metadata["{}-uri".format(system)]
        except KeyError:
            return default
        return uri

    def set_uri(self, system, value):
        key = "{}-uri".format(system)
        self.metadata[key] = value


class Document(object):
    """An HTML document noted as ``content`` on the instance,
    which can contain ``Resource`` instances.
    """
    media_type = 'application/xhtml+xml'

    def __init__(self, id, data, metadata=None, resources=None,
                 reference_resolver=None):
        self.id = id
        self._xml = None
        if hasattr(data, 'read'):
            self.content = utf8(data.read())
        else:
            self.content = utf8(data)
        self._references = _parse_references(self._xml)
        self.metadata = utf8(metadata or {})
        self.resources = resources or []

    @property
    def html(self):
        html = lxml.html.builder.HTML(
            lxml.html.fragment_fromstring(self.content, 'body'))
        return utf8(lxml.html.tostring(html, method='xml'))

    def _content__get(self):
        """Produce the content from the data.
        This is used to write out reference changes that may have
        taken place.
        """
        string_types = (type(u''), type(b''))
        # Unwrap the xml.
        content = [isinstance(node, string_types) and node or etree.tostring(node)
                   for node in self._xml.xpath('node()')]
        return ''.join(utf8(content))

    def _content__set(self, value):
        self._xml = lxml.html.fragment_fromstring(value, 'div')
        # reload the references after a content update
        self._references = _parse_references(self._xml)

    def _content__del(self):
        self._xml = etree.Element('div')

    content = property(_content__get,
                       _content__set,
                       _content__del,
                       _content__get.__doc__)

    @property
    def ident_hash(self):
        if self.id is not None:
            args = [self.id]
            version = self.metadata.get('version')
            if version is not None:
                args.append(version)
            value = '@'.join(args)
        else:
            value = None
        return value

    def get_uri(self, system, default=None):
        try:
            uri = self.metadata["{}-uri".format(system)]
        except KeyError:
            return default
        return uri

    def set_uri(self, system, value):
        key = "{}-uri".format(system)
        self.metadata[key] = value

    @property
    def references(self):
        """Reference points in the document.
        These could be resources, other documents, external links, etc.
        """
        if self._references is None:
            return []
        return self._references


class DocumentPointer(object):
    media_type = 'application/xhtml+xml'

    def __init__(self, ident_hash, metadata=None):
        self.ident_hash = ident_hash
        self.id = ident_hash
        self.metadata = metadata is not None and metadata or {}

    @classmethod
    def from_uri(cls, uri):
        parts = urlparse(uri)
        split_path = parts.path.split('/')
        ident_hash = split_path[-1]
        return cls(ident_hash)


class Resource(object):
    """A binary object used within the context of the ``Document``.
    It is typically referenced within the documents HTML content.
    """

    def __init__(self, id, data, media_type, filename=None):
        self.id = id
        if not isinstance(data, io.BytesIO):
            raise ValueError("Data must be an io.BytesIO instance. "
                             "'{}' was given.".format(type(data)))
        self._data = data
        self.media_type = media_type

        self._hash = hashlib.new(RESOURCE_HASH_TYPE,
                                 self._data.read()).hexdigest()
        if not filename:
            # Create a filename from the hash and media-type.
            filename = "{}.{}".format(
                self._hash, mimetypes.guess_extension(self.media_type))
        self.filename = filename

        self._data.seek(0)

    @property
    def hash(self):
        return self._hash

    @contextmanager
    def open(self):
        self._data.seek(0)
        yield self._data
        self._data.seek(0)



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2013, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
import os
import unittest

from lxml import etree


here = os.path.abspath(os.path.dirname(__file__))
TEST_DATA_DIR = os.path.join(here, 'data')


class HTMLParsingTestCase(unittest.TestCase):
    maxDiff = None

    def test_metadata_parsing(self):
        """Verify the parsing of metadata from an HTML document."""
        html_doc_filepath = os.path.join(
            TEST_DATA_DIR, 'book', 'content',
            'e78d4f90-e078-49d2-beac-e95e8be70667@3.xhtml')
        from ..html_parsers import parse_metadata
        with open(html_doc_filepath, 'r') as fb:
            html = etree.parse(fb)
            metadata = parse_metadata(html)
        summary = b"""<div xmlns="http://www.w3.org/1999/xhtml" xmlns:bib="http://bibtexml.sf.net/" xmlns:data="http://dev.w3.org/html5/spec/#custom" xmlns:epub="http://www.idpf.org/2007/ops" xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:lrmi="http://lrmi.net/the-specification" class="description" itemprop="description" data-type="description">\n        By the end of this section, you will be able to: \n        <ul class="list">\n          <li class="item">Drive a car</li>\n          <li class="item">Purchase a watch</li>\n          <li class="item">Wear funny hats</li>\n          <li class="item">Eat cake</li>\n        </ul>\n      </div>\n\n      """
        expected_metadata = {
            'summary': summary,
            'authors': [
                {'id': 'https://github.com/marknewlyn',
                 'name': 'Mark Horner',
                 'type': 'github-id'},
                {'id': 'https://cnx.org/member_profile/sarblyth',
                 'name': 'Sarah Blyth',
                 'type': 'cnx-id'},
                {'id': 'https://example.org/profiles/charrose',
                 'name': 'Charmaine St. Rose',
                 'type': 'openstax-id'}],
            'copyright_holders': [
                {'id': 'https://cnx.org/member_profile/ream',
                 'name': 'Ream',
                 'type': 'cnx-id'}],
            'created': '2013/03/19 15:01:16 -0500',
            'editors': [{'id': None, 'name': 'I. M. Picky', 'type': None}],
            'illustrators': [{'id': None, 'name': 'Francis Hablar',
                              'type': None}],
            'keywords': ['South Africa'],
            'license_text': 'CC-By 4.0',
            'license_url': 'http://creativecommons.org/licenses/by/4.0/',
            'publishers': [{'id': None, 'name': 'Ream', 'type': None}],
            'revised': '2013/06/18 15:22:55 -0500',
            'subjects': ['Science and Mathematics'],
            'title': 'Document One of Infinity',
            'translators': [{'id': None, 'name': 'Francis Hablar',
                             'type': None}],
            'cnx-archive-uri': 'e78d4f90-e078-49d2-beac-e95e8be70667',
            'derived_from_uri': 'http://example.org/contents/id@ver',
            'derived_from_title': 'Wild Grains and Warted Feet',
            }
        self.assertEqual(metadata, expected_metadata)



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2013, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
try:
    import html.parser as HTMLParser
except:
    import HTMLParser
import mimetypes
import os
import io
import tempfile
import shutil
import random
import unittest
try:
    from unittest import mock
except ImportError:
    import mock

from lxml import etree


here = os.path.abspath(os.path.dirname(__file__))
TEST_DATA_DIR = os.path.join(here, 'data')


def unescape(html):
    p = HTMLParser.HTMLParser()
    if isinstance(html, bytes):
        html = html.decode('utf-8')
    return p.unescape(html)


def random_extension(*args, **kwargs):
    # mimetypes.guess_extension can return any of the values in
    # mimetypes.guess_all_extensions.  it depends on the system.
    # we're using this to make sure our code is robust enough to handle the
    # different possible extensions
    exts = mimetypes.guess_all_extensions(*args, **kwargs)
    return random.choice(exts)


class AdaptationTestCase(unittest.TestCase):
    maxDiff = None

    def make_package(self, file):
        from ..epub import Package
        return Package.from_file(file)

    def make_item(self, file, **kwargs):
        from ..epub import Item
        return Item.from_file(file, **kwargs)

    def test_to_binder(self):
        """Adapts a ``Package`` to a ``BinderItem``.
        Binders are native object representations of data,
        while the Package is merely a representation of EPUB structure.
        """
        # Easiest way to test this is using the ``model_to_tree`` utility
        # to analyze the structural equality.
        package_filepath = os.path.join(
            TEST_DATA_DIR, 'book',
            "9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6.opf")
        package = self.make_package(package_filepath)
        expected_tree = {
            'id': None,
            'title': 'Book of Infinity',
            'contents': [
                {'id': 'subcol',
                 'title': 'Part One',
                 'contents': [
                     {
                      'contents': [
                          {'id': 'e78d4f90-e078-49d2-beac-e95e8be70667', 'title': 'Document One'}],
                             'id': 'subcol',
                             'title': 'Chapter One'},
                     {'id': 'subcol',
                      'title': 'Chapter Two',
                      'contents': [{'id': 'e78d4f90-e078-49d2-beac-e95e8be70667',
                                    'title': 'Document One (again)'}],
                      }]},
                {'id': 'subcol',
                 'title': 'Part Two',
                 'contents': [
                     {'id': 'subcol',
                      'title': 'Chapter Three',
                      'contents': [
                          {'id': 'e78d4f90-e078-49d2-beac-e95e8be70667',
                           'title': 'Document One (...and again)'}]
                      }]}]}

        from ..adapters import adapt_package
        binder = adapt_package(package)

        # This checks the binder structure, and only taps at the documents.
        from ..models import model_to_tree
        tree = model_to_tree(binder)
        self.assertEqual(tree, expected_tree)
        self.assertEqual(package.metadata['publication_message'], u'Nueva Versión')

    def test_to_translucent_binder(self):
        """Adapts a ``Package`` to a ``TranslucentBinder``.
        Translucent binders are native object representations of data,
        while the Package is merely a representation of EPUB structure.
        Furthermore, translucent binders are non-persistable objects,
        that contain the same behavior as binders, but lack metadata
        and material. They can be thought of as a protective sheath that
        is invisible, yet holds the contents together.
        """
        # Easiest way to test this is using the ``model_to_tree`` utility
        # to analyze the structural equality.
        package_filepath = os.path.join(
            TEST_DATA_DIR, 'loose-pages', "faux.opf")
        package = self.make_package(package_filepath)
        expected_tree = {
            'id': 'subcol',
            'title': "Loose Pages",
            'contents': [{'id': None, 'title': 'Yummy'},
                         {'id': None, 'title': 'Da bomb'},
                         {'id': 'pointer@1', 'title': 'Pointer'}],
            }

        from ..adapters import adapt_package
        binder = adapt_package(package)

        # This checks the binder structure, and only taps at the documents.
        from ..models import model_to_tree
        tree = model_to_tree(binder)
        self.assertEqual(tree, expected_tree)

    def test_to_document_wo_resources_o_references(self):
        """Adapts an ``Item`` to a ``DocumentItem``.
        Documents are native object representations of data,
        while the Item is merely a representation of an item
        in the EPUB structure.
        We are specifically testing for metadata parsing and
        resource discovery.
        """
        item_filepath = os.path.join(
            TEST_DATA_DIR, 'loose-pages', 'content',
            "fig-bush.xhtml")
        item = self.make_item(item_filepath, media_type='application/xhtml+xml')

        package = mock.Mock()
        # This would not typically be called outside the context of
        # a package, but in the case of a scoped test we use it.
        from ..adapters import adapt_item
        document = adapt_item(item, package)

        # Check the document metadata
        expected_metadata = {
            u'authors': [{u'id': u'https://github.com/marknewlyn',
                          u'name': u'Mark Horner',
                          u'type': u'github-id'},
                         {u'id': u'https://cnx.org/member_profile/sarblyth',
                          u'name': u'Sarah Blyth',
                          u'type': u'cnx-id'},
                         {u'id': u'https://example.org/profiles/charrose',
                          u'name': u'Charmaine St. Rose',
                          u'type': u'openstax-id'}],
            u'copyright_holders': [
                {u'id': u'https://cnx.org/member_profile/ream',
                 u'name': u'Ream',
                 u'type': u'cnx-id'}],
            u'created': u'2013/03/19 15:01:16 -0500',
            u'editors': [{u'id': None, u'name': u'I. M. Picky',
                          u'type': None}],
            u'illustrators': [{u'id': None, u'name': u'Francis Hablar',
                               u'type': None}],
            u'keywords': [u'South Africa'],
            u'license_text': u'CC-By 4.0',
            u'license_url': u'http://creativecommons.org/licenses/by/4.0/',
            u'publishers': [{u'id': None, u'name': u'Ream', u'type': None}],
            u'revised': u'2013/06/18 15:22:55 -0500',
            u'subjects': [u'Science and Mathematics'],
            u'summary': u'<div xmlns="http://www.w3.org/1999/xhtml" xmlns:bib="http://bibtexml.sf.net/" xmlns:data="http://dev.w3.org/html5/spec/#custom" xmlns:epub="http://www.idpf.org/2007/ops" xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:lrmi="http://lrmi.net/the-specification" class="description" itemprop="description" data-type="description">\n        By the end of this section, you will be able to: \n        <ul class="list">\n          <li class="item">Drive a car</li>\n          <li class="item">Purchase a watch</li>\n          <li class="item">Wear funny hats</li>\n          <li class="item">Eat cake</li>\n        </ul>\n      </div>\n\n      ',
            u'title': u'Document One of Infinity',
            u'translators': [{u'id': None, u'name': u'Francis Hablar',
                              u'type': None}],
            u'derived_from_uri': u'http://example.org/contents/id@ver',
            u'derived_from_title': u'Wild Grains and Warted Feet',
            }
        self.assertEqual(expected_metadata, document.metadata)

        # Check the document uri lookup
        uri = document.get_uri('cnx-archive')
        self.assertEqual(None, uri)

        # Check resource discovery.
        self.assertEqual([], document.references)

    def test_to_document_w_resources(self):
        """Adapts an ``Item`` to a ``DocumentItem``.
        Documents are native object representations of data,
        while the Item is merely a representation of an item
        in the EPUB structure.
        We are specifically testing for reference parsing and
        resource discovery.
        """
        content_filepath = os.path.join(
            TEST_DATA_DIR, 'loose-pages', 'content',
            "fig-bush.xhtml")
        file_pointer, item_filepath = tempfile.mkstemp()
        internal_uri = "../resources/openstax.png"
        with open(content_filepath, 'r') as fb:
            xml = etree.parse(fb)
            body = xml.xpath(
                '//xhtml:body',
                namespaces={'xhtml': "http://www.w3.org/1999/xhtml"})[0]
            elm = etree.SubElement(body, "img")
            elm.set('src', internal_uri)
        with open(item_filepath, 'wb') as fb:
            fb.write(etree.tostring(xml))
        item = self.make_item(item_filepath, media_type='application/xhtml+xml')

        package = mock.Mock()
        # This would not typically be called outside the context of
        # a package, but in the case of a scoped test we use it.

        resource_filepath = os.path.join(TEST_DATA_DIR, 'loose-pages',
                                         'resources', 'openstax.png')
        from ..models import Resource
        package.grab_by_name.side_effect = [
            self.make_item(resource_filepath, media_type='image/png'),
            ]
        from ..adapters import adapt_item
        document = adapt_item(item, package)

        # Check resource discovery.
        self.assertEqual([internal_uri],
                         [ref.uri for ref in document.references])
        # Check the resource was discovered.
        self.assertEqual(['openstax.png'],
                         [res.id for res in document.resources])
        # Check that the reference is bound to the resource
        ref = list(document.references)[0]
        res = list(document.resources)[0]
        self.assertEqual(ref._bound_model, res)

    def test_to_document_pointer(self):
        """Adapts an ``Item`` to a ``DocumentPointerItem``.
        Documents are native object representations of data,
        while the Item is merely a representation of an item
        in the EPUB structure.
        """
        item_filepath = os.path.join(
                TEST_DATA_DIR, 'loose-pages', 'content',
                'pointer.xhtml')

        package = mock.Mock()
        item = self.make_item(item_filepath, media_type='application/xhtml+xml')

        from ..adapters import adapt_item, DocumentPointerItem
        pointer = adapt_item(item, package)

        self.assertEqual(type(pointer), DocumentPointerItem)
        self.assertEqual(pointer.ident_hash, 'pointer@1')
        self.assertEqual(pointer.metadata['title'], 'Pointer')


@mock.patch('mimetypes.guess_extension', new=random_extension)
class ModelsToEPUBTestCase(unittest.TestCase):

    def test_loose_pages_wo_resources(self):
        """Create a publication EPUB from a loose set of pages."""
        from ..models import TranslucentBinder, Document
        binder = TranslucentBinder(metadata={'title': "Kraken"})

        base_metadata = {
            'publishers': [],
            'created': '2013/03/19 15:01:16 -0500',
            'revised': '2013/06/18 15:22:55 -0500',
            'authors': [
                {'type': 'cnx-id',
                 'name': 'Sponge Bob',
                 'id': 'sbob'}],
            'editors': [],
            'copyright_holders': [],
            'illustrators': [],
            'subjects': ['Science and Mathematics'],
            'translators': [],
            'keywords': ['Bob', 'Sponge', 'Rock'],
            'title': "Goofy Goober Rock",
            'license_text': 'CC-By 4.0',
            'license_url': 'http://creativecommons.org/licenses/by/4.0/',
            'summary': "<p>summary</p>",
            'version': 'draft',
            }

        # Build test documents
        metadata = base_metadata.copy()
        metadata.update({'title': "entrée"})
        binder.append(Document('ingress', io.BytesIO(b'<p>Hello.</p>'),
                               metadata=metadata))
        metadata = base_metadata.copy()
        metadata.update({'title': "egress"})
        binder.append(Document('egress', io.BytesIO(u'<p>hüvasti.</p>'.encode('utf-8')),
                               metadata=metadata))

        # Call the target.
        fs_pointer, epub_filepath = tempfile.mkstemp('.epub')
        self.addCleanup(os.remove, epub_filepath)
        from ..adapters import make_publication_epub
        with open(epub_filepath, 'wb') as epub_file:
            make_publication_epub(binder, 'krabs', '$.$', epub_file)

        # Verify the results.
        epub_path = tempfile.mkdtemp('-epub')
        self.addCleanup(shutil.rmtree, epub_path)
        from ..epub import unpack_epub
        unpack_epub(epub_filepath, epub_path)

        # Because a TranslucentBinder doesn't has an id of ``None``,
        # we uniquely create one using the object's hash.
        binder_hash = str(hash(binder))
        opf_filename = "{}.opf".format(binder_hash)

        # Check filenames, generated by id and media-type.
        self.assertEqual(
            [opf_filename, 'META-INF', 'contents', 'mimetype'],
            sorted(os.listdir(epub_path)))
        filenames = sorted(os.listdir(os.path.join(epub_path, 'contents')))
        self.assertEqual(
            [binder_hash, 'egress@draft', 'ingress@draft'],
            [os.path.splitext(filename)[0] for filename in filenames])
        self.assertEqual(
            ['application/xhtml+xml', 'application/xhtml+xml', 'application/xhtml+xml'],
            [mimetypes.guess_type(filename)[0] for filename in filenames])
        navdoc_filename, egress_filename, ingress_filename = filenames

        # Check the opf file
        with open(os.path.join(epub_path, opf_filename)) as f:
            opf = unescape(f.read())
        self.assertTrue(u'<dc:publisher>krabs</dc:publisher>' in opf)
        self.assertTrue(u'<meta property="publicationMessage">$.$</meta>' in opf)

        # Check the nav
        with open(os.path.join(epub_path, 'contents', navdoc_filename)) as f:
            nav = unescape(f.read())
        expected_nav = (
                u'<nav id="toc"><ol><li>'
                u'<a href="{}">entrée</a>'
                u'</li><li>'
                u'<a href="{}">egress</a>'
                u'</li></ol></nav>'.format(ingress_filename, egress_filename))
        self.assertTrue(expected_nav in nav)

        # Check that translucent is set
        self.assertTrue('<span data-type="binding" data-value="translucent" />' in nav)

        # Check the title and content
        self.assertTrue('<title>Kraken</title>' in nav)
        with open(os.path.join(epub_path, 'contents', egress_filename)) as f:
            egress = unescape(f.read())
        self.assertTrue('<title>egress</title>' in egress)
        self.assertTrue(u'<p>hüvasti.</p>' in egress)

        # Adapt epub back to documents and binders
        from cnxepub import EPUB
        from cnxepub.adapters import adapt_package
        from cnxepub.models import flatten_model
        epub = EPUB.from_file(epub_path)
        self.assertEqual(len(epub), 1)
        binder = adapt_package(epub[0])
        self.assertEqual(len(list(flatten_model(binder))), 3)

    def test_loose_pages_w_resources(self):
        """Create a publication EPUB from a loose set of pages."""
        from ..models import TranslucentBinder, Document, Resource
        binder = TranslucentBinder(metadata={'title': "Kraken"})

        base_metadata = {
            'publishers': [],
            'created': '2013/03/19 15:01:16 -0500',
            'revised': '2013/06/18 15:22:55 -0500',
            'authors': [
                {'type': 'cnx-id',
                 'name': 'Sponge Bob',
                 'id': 'sbob'}],
            'editors': [],
            'copyright_holders': [],
            'illustrators': [],
            'subjects': ['Science and Mathematics'],
            'translators': [],
            'keywords': ['Bob', 'Sponge', 'Rock'],
            'title': "Goofy Goober Rock",
            'license_text': 'CC-By 4.0',
            'license_url': 'http://creativecommons.org/licenses/by/4.0/',
            'summary': "<p>summary</p>",
            'version': 'draft',
            }

        # Build test documents
        metadata = base_metadata.copy()
        metadata.update({'title': "entrée"})
        binder.append(Document('ingress', io.BytesIO(
            b'<p><a href="http://cnx.org/">Hello.</a></p>'),
                               metadata=metadata))
        metadata = base_metadata.copy()
        metadata.update({'title': "egress"})
        with open(os.path.join(TEST_DATA_DIR, '1x1.jpg'), 'rb') as f:
            jpg = Resource('1x1.jpg', io.BytesIO(f.read()), 'image/jpeg')
        binder.append(Document('egress', io.BytesIO(
            u'<p><img src="1x1.jpg" />hüvasti.</p>'.encode('utf-8')),
                               metadata=metadata,
                               resources=[jpg]))

        # Call the target.
        fs_pointer, epub_filepath = tempfile.mkstemp('.epub')
        self.addCleanup(os.remove, epub_filepath)
        from ..adapters import make_publication_epub
        with open(epub_filepath, 'wb') as epub_file:
            make_publication_epub(binder, 'krabs', '$.$', epub_file)

        # Verify the results.
        epub_path = tempfile.mkdtemp('-epub')
        self.addCleanup(shutil.rmtree, epub_path)
        from ..epub import unpack_epub
        unpack_epub(epub_filepath, epub_path)

        # Because a TranslucentBinder doesn't has an id of ``None``,
        # we uniquely create one using the object's hash.
        binder_hash = str(hash(binder))
        opf_filename = "{}.opf".format(binder_hash)

        # Check filenames, generated by id and media-type.
        self.assertEqual(
            [opf_filename, 'META-INF', 'contents', 'mimetype', 'resources'],
            sorted(os.listdir(epub_path)))
        filenames = sorted(os.listdir(os.path.join(epub_path, 'contents')))
        self.assertEqual(
            [binder_hash, 'egress@draft', 'ingress@draft'],
            [os.path.splitext(filename)[0] for filename in filenames])
        self.assertEqual(
            ['application/xhtml+xml', 'application/xhtml+xml', 'application/xhtml+xml'],
            [mimetypes.guess_type(filename)[0] for filename in filenames])
        self.assertEqual(os.listdir(os.path.join(epub_path, 'resources')),
            ['1x1.jpg'])
        navdoc_filename, egress_filename, ingress_filename = filenames

        # Check the opf file
        with open(os.path.join(epub_path, opf_filename)) as f:
            opf = unescape(f.read())
        self.assertTrue(u'<dc:publisher>krabs</dc:publisher>' in opf)
        self.assertTrue(u'<meta property="publicationMessage">$.$</meta>' in opf)

        # Check the nav
        with open(os.path.join(epub_path, 'contents', navdoc_filename)) as f:
            nav = unescape(f.read())
        expected_nav = (
                u'<nav id="toc"><ol><li>'
                u'<a href="{}">entrée</a>'
                u'</li><li>'
                u'<a href="{}">egress</a>'
                u'</li></ol></nav>'.format(ingress_filename, egress_filename))
        self.assertTrue(expected_nav in nav)

        # Check that translucent is set
        self.assertTrue('<span data-type="binding" data-value="translucent" />' in nav)

        # Check the title and content
        self.assertTrue('<title>Kraken</title>' in nav)
        with open(os.path.join(epub_path, 'contents', egress_filename)) as f:
            egress = unescape(f.read())
        self.assertTrue('<title>egress</title>' in egress)
        self.assertFalse('<span data-type="cnx-archive-uri"' in egress)
        self.assertTrue(u'<p><img src="../resources/1x1.jpg"/>hüvasti.</p>' in egress)

        # Adapt epub back to documents and binders
        from cnxepub import EPUB
        from cnxepub.adapters import adapt_package
        from cnxepub.models import flatten_model
        epub = EPUB.from_file(epub_path)
        self.assertEqual(len(epub), 1)
        binder = adapt_package(epub[0])
        self.assertEqual(len(list(flatten_model(binder))), 3)

    def test_binder(self):
        """Create an EPUB from a binder with a few documents."""
        from ..models import Binder, Document, DocumentPointer
        binder_name = 'rock'
        binder = Binder(binder_name, metadata={'title': "Kraken (Nueva Versión)"})

        base_metadata = {
            'publishers': [],
            'created': '2013/03/19 15:01:16 -0500',
            'revised': '2013/06/18 15:22:55 -0500',
            'authors': [
                {'type': 'cnx-id',
                 'name': 'Sponge Bob',
                 'id': 'sbob'}],
            'editors': [],
            'copyright_holders': [],
            'illustrators': [],
            'subjects': ['Science and Mathematics'],
            'translators': [],
            'keywords': ['Bob', 'Sponge', 'Rock'],
            'title': "Goofy Goober Rock",
            'license_text': 'CC-By 4.0',
            'license_url': 'http://creativecommons.org/licenses/by/4.0/',
            'summary': "<p>summary</p>",
            'version': 'draft',
            }

        # Build test documents
        metadata = base_metadata.copy()
        metadata.update({
            'title': "entrée",
            'derived_from_uri': 'http://cnx.org/contents/dd68a67a-11f4-4140-a49f-b78e856e2262@1',
            'derived_from_title': "Taking Customers' Orders",
            })
        binder.append(Document('ingress', io.BytesIO(b'<p>Hello.</p>'),
                               metadata=metadata))
        metadata = base_metadata.copy()
        metadata.update({'title': "egress",
            'cnx-archive-uri': 'e78d4f90-e078-49d2-beac-e95e8be70667'})
        binder.append(Document('egress', io.BytesIO(u'<p>hüvasti.</p>'.encode('utf-8')),
                               metadata=metadata))
        binder.append(DocumentPointer('pointer@1', {'title': 'Pointer',
            'cnx-archive-uri': 'pointer@1',
            'url': 'http://cnx.org/contents/pointer@1'}))

        # Call the target.
        fs_pointer, epub_filepath = tempfile.mkstemp('.epub')
        self.addCleanup(os.remove, epub_filepath)
        from ..adapters import make_publication_epub
        with open(epub_filepath, 'wb') as epub_file:
            make_publication_epub(binder, 'krabs', '$.$', epub_file)

        # Verify the results.
        epub_path = tempfile.mkdtemp('-epub')
        self.addCleanup(shutil.rmtree, epub_path)
        from ..epub import unpack_epub
        unpack_epub(epub_filepath, epub_path)

        opf_filename = "{}.opf".format(binder_name)

        # Check filenames, generated by id and media-type.
        self.assertEqual(
            ['META-INF', 'contents', 'mimetype', opf_filename],
            sorted(os.listdir(epub_path)))
        filenames = sorted(os.listdir(os.path.join(epub_path, 'contents')))
        self.assertEqual(
            ['egress@draft', 'ingress@draft', 'pointer@1', binder_name],
            [os.path.splitext(filename)[0] for filename in filenames])
        self.assertEqual(
            ['application/xhtml+xml', 'application/xhtml+xml',
                'application/xhtml+xml', 'application/xhtml+xml'],
            [mimetypes.guess_type(filename)[0] for filename in filenames])
        egress_filename, ingress_filename, pointer_filename, navdoc_filename = filenames

        # Check the opf file
        with open(os.path.join(epub_path, opf_filename)) as f:
            opf = unescape(f.read())
        self.assertTrue(u'<dc:publisher>krabs</dc:publisher>' in opf)
        self.assertTrue(u'<meta property="publicationMessage">$.$</meta>' in opf)

        # Check the nav
        with open(os.path.join(epub_path, 'contents', navdoc_filename)) as f:
            nav = unescape(f.read())
        expected_nav = (
                u'<nav id="toc"><ol><li>'
                u'<a href="{}">entrée</a>'
                u'</li><li>'
                u'<a href="{}">egress</a>'
                u'</li><li>'
                u'<a href="{}">Pointer</a>'
                u'</li></ol></nav>'.format(ingress_filename, egress_filename,
                    pointer_filename))
        self.assertTrue(expected_nav in nav)

        # Check that translucent is not set
        self.assertFalse('<span data-type="binding" data-value="translucent" />' in nav)

        # Check the title and content
        self.assertTrue(u'<title>Kraken (Nueva Versión)</title>' in nav)
        with open(os.path.join(epub_path, 'contents', egress_filename)) as f:
            egress = unescape(f.read())
        with open(os.path.join(epub_path, 'contents', ingress_filename)) as f:
            ingress = unescape(f.read())
        self.assertTrue('<title>egress</title>' in egress)
        self.assertTrue('<span data-type="cnx-archive-uri" '
                'data-value="e78d4f90-e078-49d2-beac-e95e8be70667" />' in egress)
        self.assertTrue(u'<p>hüvasti.</p>' in egress)
        self.assertFalse('Derived from:' in egress)
        self.assertTrue('Derived from:' in ingress)
        self.assertTrue('http://cnx.org/contents/dd68a67a-11f4-4140-a49f-b78e856e2262@1' in ingress)
        self.assertTrue("Taking Customers' Orders" in ingress)

        # Check the content of the document pointer file
        with open(os.path.join(epub_path, 'contents', pointer_filename)) as f:
            pointer = unescape(f.read())
        self.assertTrue('<title>Pointer</title>' in pointer)
        self.assertTrue('<span data-type="document" data-value="pointer" />' in pointer)
        self.assertTrue('<span data-type="cnx-archive-uri" '
                'data-value="pointer@1" />' in pointer)
        self.assertTrue('<a href="http://cnx.org/contents/pointer@1">here</a>' in pointer)

        # Adapt epub back to documents and binders
        from cnxepub import EPUB
        from cnxepub.adapters import adapt_package
        from cnxepub.models import flatten_model
        epub = EPUB.from_file(epub_path)
        self.assertEqual(len(epub), 1)
        binder = adapt_package(epub[0])
        self.assertEqual(len(list(flatten_model(binder))), 4)



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2013, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
import os
import io
import json
import unittest
try:
    from unittest import mock
except ImportError:
    import mock


here = os.path.abspath(os.path.dirname(__file__))
TEST_DATA_DIR = os.path.join(here, 'data')


class TreeUtilityTestCase(unittest.TestCase):

    def make_binder(self, id=None, nodes=None, metadata=None):
        """Make a ``Binder`` instance.
        If ``id`` is not supplied, a ``FauxBinder`` is made.
        """
        from ..models import Binder, TranslucentBinder
        if id is None:
            binder = TranslucentBinder(nodes, metadata)
        else:
            binder = Binder(id, nodes, metadata)
        return binder

    def make_document(self, id, metadata={}):
        from ..models import Document
        return Document(id, io.BytesIO(b''), metadata=metadata)

    def make_document_pointer(self, ident_hash, metadata={}):
        from ..models import DocumentPointer
        return DocumentPointer(ident_hash, metadata=metadata)

    def test_binder_to_tree(self):
        binder = self.make_binder(
            '8d75ea29',
            metadata={'version': '3', 'title': "Book One"},
            nodes=[
                self.make_binder(
                    None,
                    metadata={'title': "Part One"},
                    nodes=[
                        self.make_binder(
                            None,
                            metadata={'title': "Chapter One"},
                            nodes=[
                                self.make_document(
                                    id="e78d4f90",
                                    metadata={'version': '3',
                                              'title': "Document One"})]),
                        self.make_binder(
                            None,
                            metadata={'title': "Chapter Two"},
                            nodes=[
                                self.make_document(
                                    id="3c448dc6",
                                    metadata={'version': '1',
                                              'title': "Document Two"})])]),
                self.make_binder(
                    None,
                    metadata={'title': "Part Two"},
                    nodes=[
                        self.make_binder(
                            None,
                            metadata={'title': "Chapter Three"},
                            nodes=[
                                self.make_document(
                                    id="ad17c39c",
                                    metadata={'version': '2',
                                              'title': "Document Three"})])]),
                self.make_binder(
                    None,
                    metadata={'title': "Part Three"},
                    nodes=[
                        self.make_binder(
                            None,
                            metadata={'title': "Chapter Four"},
                            nodes=[
                                self.make_document(
                                    id="7c52af05",
                                    metadata={'version': '1',
                                              'title': "Document Four"})])])])

        expected_tree = {
            'id': '8d75ea29@3',
            'contents': [
                {'id': 'subcol',
                 'contents': [
                     {'id': 'subcol',
                      'contents': [
                          {'id': 'e78d4f90@3',
                           'title': 'Document One'}],
                      'title': 'Chapter One'},
                     {'id': 'subcol',
                      'contents': [
                          {'id': '3c448dc6@1',
                           'title': 'Document Two'}],
                      'title': 'Chapter Two'}],
                 'title': 'Part One'},
                {'id': 'subcol',
                 'contents': [
                    {'id': 'subcol',
                     'contents': [
                         {'id': 'ad17c39c@2',
                          'title': 'Document Three'}],
                     'title': 'Chapter Three'}],
                 'title': 'Part Two'},
                {'id': 'subcol',
                 'contents': [
                     {'id': 'subcol',
                      'contents': [
                          {'id': '7c52af05@1',
                           'title': 'Document Four'}],
                      'title': 'Chapter Four'}],
                 'title': 'Part Three'}],
            'title': 'Book One'}

        from ..models import model_to_tree
        tree = model_to_tree(binder)
        self.assertEqual(tree, expected_tree)

    def test_flatten_model(self):
        binder = self.make_binder(
            '8d75ea29',
            metadata={'version': '3', 'title': "Book One"},
            nodes=[
                self.make_binder(
                    None,
                    metadata={'title': "Part One"},
                    nodes=[
                        self.make_binder(
                            None,
                            metadata={'title': "Chapter One"},
                            nodes=[
                                self.make_document(
                                    id="e78d4f90",
                                    metadata={'version': '3',
                                              'title': "Document One"})]),
                        self.make_binder(
                            None,
                            metadata={'title': "Chapter Two"},
                            nodes=[
                                self.make_document(
                                    id="3c448dc6",
                                    metadata={'version': '1',
                                              'title': "Document Two"})])]),
                self.make_binder(
                    None,
                    metadata={'title': "Part Two"},
                    nodes=[
                        self.make_binder(
                            None,
                            metadata={'title': "Chapter Three"},
                            nodes=[
                                self.make_document(
                                    id="ad17c39c",
                                    metadata={'version': '2',
                                              'title': "Document Three"})])])])
        expected_titles = [
            'Book One',
            'Part One',
            'Chapter One', 'Document One',
            'Chapter Two', 'Document Two',
            'Part Two',
            'Chapter Three', 'Document Three']

        from ..models import flatten_model
        titles = [m.metadata['title'] for m in flatten_model(binder)]
        self.assertEqual(titles, expected_titles)

    def test_flatten_to_documents(self):
        binder = self.make_binder(
            '8d75ea29',
            metadata={'version': '3', 'title': "Book One"},
            nodes=[
                self.make_binder(
                    None,
                    metadata={'title': "Part One"},
                    nodes=[
                        self.make_binder(
                            None,
                            metadata={'title': "Chapter One"},
                            nodes=[
                                self.make_document(
                                    id="e78d4f90",
                                    metadata={'version': '3',
                                              'title': "Document One"})]),
                        self.make_document_pointer(
                            ident_hash='844a99e5@1',
                            metadata={'title': "Pointing"}),
                        self.make_binder(
                            None,
                            metadata={'title': "Chapter Two"},
                            nodes=[
                                self.make_document(
                                    id="3c448dc6",
                                    metadata={'version': '1',
                                              'title': "Document Two"})])]),
                self.make_binder(
                    None,
                    metadata={'title': "Part Two"},
                    nodes=[
                        self.make_binder(
                            None,
                            metadata={'title': "Chapter Three"},
                            nodes=[
                                self.make_document(
                                    id="ad17c39c",
                                    metadata={'version': '2',
                                              'title': "Document Three"})])])])

        from ..models import flatten_to_documents

        # Test for default, Document only results.
        expected_titles = ['Document One', 'Document Two', 'Document Three']
        titles = [d.metadata['title'] for d in flatten_to_documents(binder)]
        self.assertEqual(titles, expected_titles)

        # Test for included DocumentPointer results.
        expected_titles = ['Document One', 'Pointing', 'Document Two',
                           'Document Three']
        titles = [d.metadata['title']
                  for d in flatten_to_documents(binder, include_pointers=True)]
        self.assertEqual(titles, expected_titles)


class ModelBehaviorTestCase(unittest.TestCase):

    def test_document_w_references(self):
        """Documents are loaded then parsed to show their
        references within the HTML content.
        """
        expected_uris = ["http://example.com/people/old-mcdonald",
                         "http://cnx.org/contents/5f3acd92@3",
                         "../resources/nyan-cat.gif"
                         ]
        content = """\
<h1> McDonald Bio </h1>
<p>There is a farmer named <a href="{}">Old McDonald</a>. Plants grow on his farm and animals live there. He himself is vegan, and so he wrote a book about <a href="{}">Vegan Farming</a>.</p>
<img src="{}"/>
<span>Ei ei O.</span>
""".format(*expected_uris)

        from ..models import Document
        document = Document('mcdonald', content)

        self.assertEqual(len(document.references), 3)
        are_external = [r.remote_type == 'external'
                        for r in document.references]
        self.assertEqual([True, True, False], are_external)
        self.assertEqual(expected_uris, [r.uri for r in document.references])

        # reload the content
        document.content = content
        # update some references
        document.references[0].uri = 'https://example.com/people/old-mcdonald'
        self.assertTrue('<a href="https://example.com/people/old-mcdonald">'
                        in document.content)

    def test_document_w_bound_references(self):
        starting_uris = ["../resources/openstax.png",
                         "m23409.xhtml",
                         ]
        content = """\
<h1>Reference replacement test-case</h1>
<p>Link to <a href="{}">a local legacy module</a>.</p>
<img src="{}"/>
<p>Fin.</p>
""".format(*starting_uris)

        from ..models import Document
        document = Document('document', content)

        self.assertEqual(len(document.references), 2)
        are_external = [r.remote_type == 'external'
                        for r in document.references]
        self.assertEqual([False, False], are_external)
        self.assertEqual(starting_uris, [r.uri for r in document.references])

        # Now bind the model to the reference.
        resource_uri_tmplt = "/resources/{}"
        resource_name = '36ad78c3'
        resource = mock.Mock()
        resource.id = resource_name
        document.references[0].bind(resource, "/resources/{}")

        expected_uris = [
            resource_uri_tmplt.format(resource_name),
            starting_uris[1],
            ]
        self.assertEqual(expected_uris, [r.uri for r in document.references])

        # And change it the resource identifier
        changed_resource_name = 'smoo.png'
        resource.id = changed_resource_name
        expected_uris = [
            resource_uri_tmplt.format(changed_resource_name),
            starting_uris[1],
            ]
        self.assertEqual(expected_uris, [r.uri for r in document.references])

    def test_document_content(self):
        with open(os.path.join(TEST_DATA_DIR,
            'fb74dc89-47d4-4e46-aac1-b8682f487bd5@1.json'),
            'r') as f:
            metadata = json.loads(f.read())
        from ..models import Document
        document = Document('document', metadata['content'])
        self.assertTrue('To demonstrate the potential of online publishing'
                in document.content)



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2013, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
from lxml import etree

from .models import TRANSLUCENT_BINDER_ID


HTML_DOCUMENT_NAMESPACES = {
    'xhtml': "http://www.w3.org/1999/xhtml",
    'epub': "http://www.idpf.org/2007/ops",
    }


def parse_navigation_html_to_tree(html, id):
    """Parse the given ``html`` (an etree object) to a tree.
    The ``id`` is required in order to assign the top-level tree id value.
    """
    xpath = lambda x: html.xpath(x, namespaces=HTML_DOCUMENT_NAMESPACES)
    try:
        value = xpath('//*[@data-type="binding"]/@data-value')[0]
        is_translucent = value == 'translucent'
    except IndexError:
        is_translucent = False
    if is_translucent:
        id = TRANSLUCENT_BINDER_ID
    tree = {'id': id,
            'title': xpath('//*[@data-type="title"]/text()')[0],
            'contents': [x for x in _nav_to_tree(xpath('//xhtml:nav')[0])]
            }
    return tree


def _nav_to_tree(root):
    """Given an etree containing a navigation document structure
    rooted from the 'nav' element, parse to a tree:
    {'id': <id>|'subcol', 'title': <title>, 'contents': [<tree>, ...]}
    """
    expath = lambda e,x: e.xpath(x, namespaces=HTML_DOCUMENT_NAMESPACES)
    for li in expath(root, 'xhtml:ol/xhtml:li'):
        classes = li.get('class', '').split()
        is_subtree = bool([e for e in li.getchildren()
                           if e.tag[e.tag.find('}')+1:] == 'ol'])
        if is_subtree:
            # It's a sub-tree and have a 'span' and 'ol'.
            yield {'id': 'subcol', # Special id...
                   'title': expath(li, 'xhtml:span/text()')[0],
                   'contents': [x for x in _nav_to_tree(li)],
                   }
        else:
            # It's a node and should only have an li.
            a = li.xpath('xhtml:a', namespaces=HTML_DOCUMENT_NAMESPACES)[0]
            yield {'id': a.get('href'), 'title': a.text}
    raise StopIteration()


def parse_metadata(html):
    """Parse metadata out of the given an etree object as ``html``."""
    parser = DocumentMetadataParser(html)
    return parser()


class DocumentMetadataParser:
    """Given a file-like object, parse out the metadata to a dictionary.
    This only parses the data. It does not validate it.
    """
    namespaces = HTML_DOCUMENT_NAMESPACES
    metadata_required_keys = (
        'title', 'license_url', 'summary',
        )
    metadata_optional_keys = (
        'created', 'revised', 'language', 'subjects', 'keywords',
        'license_text', 'editors', 'illustrators', 'translators',
        'publishers', 'copyright_holders', 'authors',
        'cnx-archive-uri', 'derived_from_uri', 'derived_from_title',
        )

    def __init__(self, elm_tree, raise_value_error=True):
       self._xml = elm_tree
       self.raise_value_error = raise_value_error

    def __call__(self):
        return self.metadata

    def parse(self, xpath, prefix=""):
        values = self._xml.xpath(prefix + xpath,
                                 namespaces=self.namespaces)
        return values

    @property
    def metadata(self):
        items = {}
        keyrings = (self.metadata_required_keys, self.metadata_optional_keys,)
        for keyring in keyrings:
            for key in keyring:
                # TODO On refactoring properties
                # raise an error on property access rather than outside of it
                # as is currently being done here.
                value = getattr(self, key.replace('-', '_'))
                if self.raise_value_error and \
                        key in self.metadata_required_keys and value is None:
                    raise ValueError(
                        "A value for '{}' could not be found.".format(key))
                elif value is None:
                    continue
                items[key] = value
        return items

    @property
    def title(self):
        items = self.parse('//xhtml:*[@data-type="title"]/text()')
        try:
            value = items[0]
        except IndexError:
            value = None
        return value

    @property
    def summary(self):
        items = self.parse('//xhtml:*[@data-type="description"]')
        try:
            value = items[0]
            value = etree.tostring(value)
        except IndexError:
            value = None
        return value

    @property
    def created(self):
        items = self.parse('//xhtml:meta[@itemprop="dateCreated"]/@content')
        try:
            value = items[0]
        except IndexError:
            value = None
        return value

    @property
    def revised(self):
        items = self.parse('//xhtml:meta[@itemprop="dateModified"]/@content')
        try:
            value = items[0]
        except IndexError:
            value = None
        return value

    @property
    def language(self):
        items = self.parse('//xhtml:*[@data-type="language"]/@content')
        try:
            value = items[0]
        except IndexError:
            value = None
        return value

    @property
    def subjects(self):
        items = self.parse('//xhtml:*[@data-type="subject"]/text()')
        return items

    @property
    def keywords(self):
        items = self.parse('//xhtml:*[@data-type="keyword"]/text()')
        return items

    @property
    def license_url(self):
        items = self.parse('//xhtml:*[@data-type="license"]/@href')
        try:
            value = items[0]
        except IndexError:
            value = None
        return value

    @property
    def license_text(self):
        items = self.parse('//xhtml:*[@data-type="license"]/text()')
        try:
            value = items[0]
        except IndexError:
            value = None
        return value

    def _parse_person_info(self, xpath):
        unordered = []
        for elm in self.parse(xpath):
            elm_id = elm.get('id', None)
            if len(elm) > 0:
                person_elm = elm[0]
                name = person_elm.text
                type_ = person_elm.get('data-type', None)
                id_ = person_elm.get('href', None)
            else:
                name = elm.text
                type_ = None
                id_ = None
            person = {'name': name, 'type': type_, 'id': id_}
            # Meta refinement allows these to be ordered.
            order = None
            refines_xpath_tmplt = """//xhtml:meta[@refines="#{}" and @property="display-seq"]/@content"""
            if elm_id is not None:
                try:
                    order = self.parse(refines_xpath_tmplt.format(elm_id))[0]
                except IndexError:
                    pass # Check for refinement failed, maintain None value.
            unordered.append((order, person,))
        ordered = sorted(unordered, key=lambda x: x[0])
        values = [x[1] for x in ordered]
        return values

    @property
    def publishers(self):
        xpath = '//xhtml:*[@data-type="publisher"]'
        return self._parse_person_info(xpath)

    @property
    def editors(self):
        xpath = '//xhtml:*[@data-type="editor"]'
        return self._parse_person_info(xpath)

    @property
    def illustrators(self):
        xpath = '//xhtml:*[@data-type="illustrator"]'
        return self._parse_person_info(xpath)

    @property
    def translators(self):
        xpath = '//xhtml:*[@data-type="translator"]'
        return self._parse_person_info(xpath)

    @property
    def copyright_holders(self):
        xpath = '//xhtml:*[@data-type="copyright-holder"]'
        return self._parse_person_info(xpath)

    @property
    def authors(self):
        xpath = '//xhtml:*[@data-type="author"]'
        return self._parse_person_info(xpath)

    @property
    def cnx_archive_uri(self):
        items = self.parse('//xhtml:*[@data-type="cnx-archive-uri"]/@data-value')
        if items:
            return items[0]

    @property
    def derived_from_uri(self):
        items = self.parse('//xhtml:*[@data-type="derived-from"]/@href')
        if items:
            return items[0]

    @property
    def derived_from_title(self):
        items = self.parse('//xhtml:*[@data-type="derived-from"]/text()')
        if items:
            return items[0]


class DocumentPointerMetadataParser(DocumentMetadataParser):
    metadata_required_keys = (
            'title', 'cnx-archive-uri', 'is_document_pointer',
            )
    metadata_optional_keys = DocumentMetadataParser.metadata_optional_keys + (
            'license_url', 'summary',
            )

    @property
    def is_document_pointer(self):
        items = self.parse('//xhtml:*[@data-type="document"]/@data-value')
        if items:
            return items[0] == 'pointer'



# -*- coding: utf-8 -*-
import sys
from setuptools import setup, find_packages


IS_PY3 = sys.version_info > (3,)


install_requires = [
    'jinja2',
    'lxml',
    ]
tests_require = [
    ]
extras_require = {
    'test': tests_require,
    }
description = "Library for building and paring Connexions' EPUBs."

if not IS_PY3:
    tests_require.append('mock')


setup(
    name='cnx-epub',
    version='0.6.0',
    author='Connexions team',
    author_email='info@cnx.org',
    url="https://github.com/connexions/cnx-epub",
    license='LGPL, See also LICENSE.txt',
    description=description,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,
    packages=find_packages(),
    include_package_data=False,
    entry_points="""\
    [console_scripts]
    """,
    test_suite='cnxepub.tests',
    zip_safe=False,
    )



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2013, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
import os
import io
import mimetypes
from copy import deepcopy

import jinja2
from lxml import etree

from .epub import EPUB, Package, Item
from .models import (
    model_to_tree, flatten_model,
    Binder, TranslucentBinder,
    Document, Resource, DocumentPointer,
    TRANSLUCENT_BINDER_ID,
    INTERNAL_REFERENCE_TYPE,
    INLINE_REFERENCE_TYPE,
    )
from .html_parsers import (parse_metadata, parse_navigation_html_to_tree,
        DocumentPointerMetadataParser)

from .data_uri import DataURI

__all__ = (
    'adapt_package', 'adapt_item',
    'make_epub', 'make_publication_epub',
    'BinderItem',
    'DocumentItem',
    )


def adapt_package(package):
    """Adapts ``.epub.Package`` to a ``BinderItem`` and cascades
    the adaptation downward to ``DocumentItem``
    and ``ResourceItem``.
    The results of this process provide the same interface as
    ``.models.Binder``, ``.models.Document`` and ``.models.Resource``.
    """
    navigation_item = package.navigation
    html = etree.parse(navigation_item.data)
    tree = parse_navigation_html_to_tree(html, navigation_item.name)
    return _node_to_model(tree, package)


def adapt_item(item, package):
    """Adapts ``.epub.Item`` to a ``DocumentItem``.

    """
    if item.media_type == 'application/xhtml+xml':
        html = etree.parse(item.data)
        metadata = DocumentPointerMetadataParser(html, raise_value_error=False)()
        item.data.seek(0)
        if metadata.get('is_document_pointer'):
            model = DocumentPointerItem(item, package)
        else:
            model = DocumentItem(item, package)
    else:
        model = Resource(item.name, item.data, item.media_type, item.name)
    return model


def make_epub(binders, file):
    """Creates an EPUB file from a binder(s)."""
    if not isinstance(binders, (list, set, tuple,)):
        binders = [binders]
    epub = EPUB([_make_package(binder) for binder in binders])
    epub.to_file(epub, file)


def make_publication_epub(binders, publisher, publication_message, file):
    """Creates an epub file from a binder(s). Also requires
    publication information, meant to be used in a EPUB publication
    request.
    """
    if not isinstance(binders, (list, set, tuple,)):
        binders = [binders]
    packages = []
    for binder in binders:
        metadata = binder.metadata
        binder.metadata = deepcopy(metadata)
        binder.metadata.update({'publisher': publisher,
                                'publication_message': publication_message})
        packages.append(_make_package(binder))
        binder.metadata = metadata
    epub = EPUB(packages)
    epub.to_file(epub, file)


def _make_package(binder):
    """Makes an ``.epub.Package`` from a  Binder'ish instance."""
    package_id = binder.id
    if package_id is None:
        package_id = hash(binder)

    package_name = "{}.opf".format(package_id)

    extensions = {}
    # Set model identifier file extensions.
    for model in flatten_model(binder):
        if isinstance(model, (Binder, TranslucentBinder,)):
            continue
        ext = mimetypes.guess_extension(model.media_type, strict=False)
        if ext is None:
            raise ValueError("Can't apply an extension to media-type '{}'." \
                             .format(modle.media_type))
        extensions[model.id] = ext
        extensions[model.ident_hash] = ext

    template_env = jinja2.Environment(trim_blocks=True, lstrip_blocks=True)
    isdict = lambda v: isinstance(v, dict)
    template = template_env.from_string(HTML_DOCUMENT,
                                        globals={'isdict': isdict})
    # Build the package item list.
    items = []
    # Build the binder as an item, specifically a navigation item.
    navigation_content =  tree_to_html(model_to_tree(binder), extensions)
    navigation_document = template.render(metadata=binder.metadata,
                                          content=navigation_content,
                                          is_translucent=binder.is_translucent)
    navigation_document_name = "{}{}".format(package_id,
            mimetypes.guess_extension('application/xhtml+xml', strict=False))
    item = Item(str(navigation_document_name),
                io.BytesIO(navigation_document.encode('utf-8')),
                'application/xhtml+xml', is_navigation=True, properties=['nav'])
    items.append(item)
    resources = {}
    pointer_template = jinja2.Template(DOCUMENT_POINTER_TEMPLATE,
                                       trim_blocks=True, lstrip_blocks=True)
    # Roll through the model list again, making each one an item.
    for model in flatten_model(binder):
        if isinstance(model, (Binder, TranslucentBinder,)):
            continue
        if isinstance(model, DocumentPointer):
            content = pointer_template.render(metadata=model.metadata)
            item = Item(''.join([model.ident_hash, extensions[model.id]]),
                        io.BytesIO(content.encode('utf-8')),
                        model.media_type)
            items.append(item)
            continue
        for resource in model.resources:
            resources[resource.id] = resource
            with resource.open() as data:
                item = Item(resource.id, data, resource.media_type)
            items.append(item)
        for reference in model.references:
            if reference.remote_type == INLINE_REFERENCE_TYPE:
                # has side effects - converts ref type to INTERNAL w/
                # appropriate uri, so need to replicate resource treatment from above
                resource = _make_resource_from_inline(reference)
                model.resources.append(resource)
                resources[resource.id] = resource
                with resource.open() as data:
                    item = Item(resource.id, data, resource.media_type)
                items.append(item)
                reference.bind(resource,'../resources/{}')

            elif reference.remote_type == INTERNAL_REFERENCE_TYPE:
                filename = os.path.basename(reference.uri)
                resource = resources.get(filename)
                if resource:
                    reference.bind(resource, '../resources/{}')

        complete_content = template.render(metadata=model.metadata,
                                           content=model.content)
        item = Item(''.join([model.ident_hash, extensions[model.id]]),
                    io.BytesIO(complete_content.encode('utf-8')),
                    model.media_type)
        items.append(item)

    # Build the package.
    package = Package(package_name, items, binder.metadata)
    return package

def _make_resource_from_inline(reference):
    """Makes an ``models.Resource`` from a ``models.Reference``
       of type INLINE. That is, a data: uri"""
    uri = DataURI(reference.uri)
    data = io.BytesIO(uri.data)
    mimetype = uri.mimetype
    res = Resource('dummy', data, mimetype)
    res.id = res.filename
    return res
    

def _make_item(model):
    """Makes an ``.epub.Item`` from
    a ``.models.Document`` or ``.models.Resource``
    """
    item = Item(model.id, model.content, model.media_type)


def _node_to_model(tree_or_item, package, parent=None,
                   lucent_id=TRANSLUCENT_BINDER_ID):
    """Given a tree, parse to a set of models"""
    if 'contents' in tree_or_item:
        # It is a binder.
        tree = tree_or_item
        if tree['id'] == lucent_id:
            binder = TranslucentBinder(metadata={'title': tree['title']})
        else:
            package_item = package.grab_by_name(tree['id'])
            binder = BinderItem(package_item, package)
        for item in tree['contents']:
            node = _node_to_model(item, package, parent=binder,
                                  lucent_id=lucent_id)
            if node.metadata['title'] != item['title']:
                binder.set_title_for_node(node, item['title'])
        result = binder
    else:
        # It is a document.
        item = tree_or_item
        package_item = package.grab_by_name(item['id'])
        result = adapt_item(package_item, package)
    if parent is not None:
        parent.append(result)
    return result


def _id_from_metadata(metadata):
    """Given an item's metadata, discover the id."""
    # FIXME Where does the system identifier come from?
    system = 'cnx-archive'
    identifier = "{}-uri".format(system)
    if identifier in metadata:
        id = metadata[identifier]
    else:
        id = None
    return id


class BinderItem(Binder):

    def __init__(self, item, package):
        self._item = item
        self._package = package
        html = etree.parse(self._item.data)
        metadata = parse_metadata(html)
        id = _id_from_metadata(metadata)
        super(BinderItem, self).__init__(id, metadata=metadata)


class DocumentPointerItem(DocumentPointer):

    def __init__(self, item, package):
        self._item = item
        self._package = package
        self._html = etree.parse(self._item.data)

        metadata = DocumentPointerMetadataParser(self._html)()
        id = _id_from_metadata(metadata)
        super(DocumentPointerItem, self).__init__(id, metadata=metadata)


class DocumentItem(Document):

    def __init__(self, item, package):
        self._item = item
        self._package = package
        self._html = etree.parse(self._item.data)

        metadata = parse_metadata(self._html)
        content_xpath = "//xhtml:body/node()[not(self::node()[@data-type='metadata'])]"
        nsmap = {'xhtml': "http://www.w3.org/1999/xhtml"}

        content = io.BytesIO(
            b''.join([isinstance(n, str) and n.encode('utf-8') or etree.tostring(n)
                      for n in self._html.xpath(content_xpath, namespaces=nsmap)]))
        id = _id_from_metadata(metadata)
        resources = None
        super(DocumentItem, self).__init__(id, content, metadata)

        # Based on the reference list, make a best effort
        # to acquire resources.
        resources = []
        for ref in self.references:
            if ref.remote_type == 'external':
                continue
            elif not ref.uri.find('../resources') >= 0:
                continue
            name = os.path.basename(ref.uri)
            try:
                resource = adapt_item(package.grab_by_name(name), package)
                ref.bind(resource, '../resources/{}')
                resources.append(resource)
            except KeyError:
                # When resources are missing, the problem is pushed off
                # to the rendering process, which will
                # raise a missing reference exception when necessary.
                pass
        self.resources = resources


# XXX Rendering shouldn't happen here.
#     Temporarily place the rendering templates and code here.

DOCUMENT_POINTER_TEMPLATE = """\
<?xml version="1.0" encoding="UTF-8"?>
<html xmlns="http://www.w3.org/1999/xhtml"
      xmlns:epub="http://www.idpf.org/2007/ops"
      xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
      xmlns:dc="http://purl.org/dc/elements/1.1/"
      xmlns:lrmi="http://lrmi.net/the-specification"
      >
  <head itemscope="itemscope"
        itemtype="http://schema.org/Book"
        >

    <title>{{ metadata['title'] }}</title>

    {# TODO Include this based on the feature being present #}
    <!-- These are for discoverability of accessible content. -->
    <meta itemprop="accessibilityFeature" content="MathML" />
    <meta itemprop="accessibilityFeature" content="LaTeX" />
    <meta itemprop="accessibilityFeature" content="alternativeText" />
    <meta itemprop="accessibilityFeature" content="captions" />
    <meta itemprop="accessibilityFeature" content="structuredNavigation" />

    {# TODO
       <meta refines="#<html-id>" property="display-seq" content="<ord>" />
     #}

  </head>
  <body xmlns:bib="http://bibtexml.sf.net/"
        xmlns:data="http://dev.w3.org/html5/spec/#custom"
        itemscope="itemscope"
        itemtype="http://schema.org/Book"
        >
    <div data-type="metadata">
      <h1 data-type="title" itemprop="name">{{ metadata['title'] }}</h1>
      <span data-type="document" data-value="pointer" />
      {% if metadata.get('cnx-archive-uri') %}
      <span data-type="cnx-archive-uri" data-value="{{ metadata['cnx-archive-uri'] }}" />
      {%- endif %}
    </div>

    <div>
      <p>
        Click <a href="{{ metadata['url'] }}">here</a> to read {{ metadata['title'] }}.
      </p>
    </div>
  </body>
</html>
"""
HTML_DOCUMENT = """\
<?xml version="1.0" encoding="UTF-8"?>
<html xmlns="http://www.w3.org/1999/xhtml"
      xmlns:epub="http://www.idpf.org/2007/ops"
      xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
      xmlns:dc="http://purl.org/dc/elements/1.1/"
      xmlns:lrmi="http://lrmi.net/the-specification"
      >
  <head itemscope="itemscope"
        itemtype="http://schema.org/Book"
        >

    <title>{{ metadata['title'] }}</title>
    <meta itemprop="inLanguage"
          data-type="language"
          content="{{ metadata['language'] }}"
          />

    {# TODO Include this based on the feature being present #}
    <!-- These are for discoverability of accessible content. -->
    <meta itemprop="accessibilityFeature" content="MathML" />
    <meta itemprop="accessibilityFeature" content="LaTeX" />
    <meta itemprop="accessibilityFeature" content="alternativeText" />
    <meta itemprop="accessibilityFeature" content="captions" />
    <meta itemprop="accessibilityFeature" content="structuredNavigation" />

    {# TODO
       <meta refines="#<html-id>" property="display-seq" content="<ord>" />
     #}

    <meta itemprop="dateCreated"
          content="{{ metadata['created'] }}"
          />
    <meta itemprop="dateModified"
          content="{{ metadata['revised'] }}"
          />
  </head>
  <body xmlns:bib="http://bibtexml.sf.net/"
        xmlns:data="http://dev.w3.org/html5/spec/#custom"
        itemscope="itemscope"
        itemtype="http://schema.org/Book"
        >
    <div data-type="metadata">
      <h1 data-type="title" itemprop="name">{{ metadata['title'] }}</h1>
      {% if is_translucent %}
      <span data-type="binding" data-value="translucent" />
      {%- endif %}
      {% if metadata.get('cnx-archive-uri') %}
      <span data-type="cnx-archive-uri" data-value="{{ metadata['cnx-archive-uri'] }}" />
      {%- endif %}

      <div class="authors">
        By: 
        {% for author in metadata['authors'] -%}
          <span id="{{ '{}-{}'.format('author', loop.index) }}"
                itemscope="itemscope"
                itemtype="http://schema.org/Person"
                itemprop="author"
                data-type="author"
                >
            <a href="{{ author['id'] }}"
               itemprop="url"
               data-type="{{ author['type'] }}"
               >{{ author['name'] }}</a>
          </span>{% if not loop.last %}, {% endif %}
        {%- endfor %}

        Edited by: 
        {% set person_type = 'editor' %}
        {% set person_itemprop_name = 'editor' %}
        {% set person_key = 'editors' %}
        {% for person in metadata[person_key] -%}
          {% if isdict(person) %}
            <span id="{{ '{}-{}'.format(person_type, loop.index) }}"
                  itemscope="itemscope"
                  itemtype="http://schema.org/Person"
                  itemprop="{{ person_type }}"
                  data-type="{{ person_type }}"
                  >
              <a href="{{ person['id'] }}"
                 itemprop="url"
                 data-type="{{ person['type'] }}"
                 >{{ person['name'] }}</a>
            </span>{% if not loop.last %}, {% endif %}
          {% else %}
            <span itemprop="{{ person_type }}"
                  data-type="{{ person_type }}"
              >person</span>{% if not loop.last %}, {% endif %}
          {% endif %}
        {%- endfor %}

        Illustrated by: 
        {% set person_type = 'illustrator' %}
        {% set person_itemprop_name = 'illustrator' %}
        {% set person_key = 'illustrators' %}
        {% for person in metadata[person_key] -%}
          {% if isdict(person) %}
            <span id="{{ '{}-{}'.format(person_type, loop.index) }}"
                  itemscope="itemscope"
                  itemtype="http://schema.org/Person"
                  itemprop="{{ person_type }}"
                  data-type="{{ person_type }}"
                  >
              <a href="{{ person['id'] }}"
                 itemprop="url"
                 data-type="{{ person['type'] }}"
                 >{{ person['name'] }}</a>
            </span>{% if not loop.last %}, {% endif %}
          {% else %}
            <span itemprop="{{ person_type }}"
                  data-type="{{ person_type }}"
              >person</span>{% if not loop.last %}, {% endif %}
          {% endif %}
        {%- endfor %}

        Translated by: 
        {% set person_type = 'translator' %}
        {% set person_itemprop_name = 'contributor' %}
        {% set person_key = 'translators' %}
        {% for person in metadata[person_key] -%}
          {% if isdict(person) %}
            <span id="{{ '{}-{}'.format(person_type, loop.index) }}"
                  itemscope="itemscope"
                  itemtype="http://schema.org/Person"
                  itemprop="{{ person_type }}"
                  data-type="{{ person_type }}"
                  >
              <a href="{{ person['id'] }}"
                 itemprop="url"
                 data-type="{{ person['type'] }}"
                 >{{ person['name'] }}</a>
            </span>{% if not loop.last %}, {% endif %}
          {% else %}
            <span itemprop="{{ person_type }}"
                  data-type="{{ person_type }}"
              >person</span>{% if not loop.last %}, {% endif %}
          {% endif %}
        {%- endfor %}

      </div>

      <div class="publishers">
        Published By: 
        {% set person_type = 'publisher' %}
        {% set person_itemprop_name = 'publisher' %}
        {% set person_key = 'publishers' %}
        {% for person in metadata[person_key] -%}
          {% if isdict(person) %}
            <span id="{{ '{}-{}'.format(person_type, loop.index) }}"
                  itemscope="itemscope"
                  itemtype="http://schema.org/Person"
                  itemprop="{{ person_type }}"
                  data-type="{{ person_type }}"
                  >
              <a href="{{ person['id'] }}"
                 itemprop="url"
                 data-type="{{ person['type'] }}"
                 >{{ person['name'] }}</a>
            </span>{% if not loop.last %}, {% endif %}
          {% else %}
            <span itemprop="{{ person_type }}"
                  data-type="{{ person_type }}"
              >person</span>{% if not loop.last %}, {% endif %}
          {% endif %}
        {%- endfor %}
      </div>

      {% if metadata.get('derived_from_uri') %}
      <div class="derived-from">
        Derived from:
        <a href="{{ metadata['derived_from_uri'] }}"
           itemprop="isDerivedFromURL"
           data-type="derived-from"
           >{{ metadata['derived_from_title'] }}</a>
      </div>
      {%- endif %}

      <div class="permissions">
        {% if metadata['copyright_holders'] %}
        <p class="copyright">
          Copyright:
          {% set person_type = 'copyright-holder' %}
          {% set person_itemprop_name = 'copyrightHolder' %}
          {% set person_key = 'copyright_holders' %}
          {% for person in metadata[person_key] -%}
            {% if isdict(person) %}
              <span id="{{ '{}-{}'.format(person_type, loop.index) }}"
                    itemscope="itemscope"
                    itemtype="http://schema.org/Person"
                    itemprop="{{ person_type }}"
                    data-type="{{ person_type }}"
                    >
                <a href="{{ person['id'] }}"
                   itemprop="url"
                   data-type="{{ person['type'] }}"
                   >{{ person['name'] }}</a>
              </span>{% if not loop.last %}, {% endif %}
            {% else %}
              <span itemprop="{{ person_type }}"
                    data-type="{{ person_type }}"
                >person</span>{% if not loop.last %}, {% endif %}
            {% endif %}
          {%- endfor %}
        </p>
        {% endif %}
        <p class="license">
          Licensed:
          <a href="{{ metadata['license_url'] }}"
             itemprop="dc:license,lrmi:useRightsURL"
             data-type="license"
             >{{ metadata['license_text'] }}</a>
        </p>
      </div>

      <div class="description"
           itemprop="description"
           data-type="description"
           >
        {{ metadata['summary'] }}
      </div>

      {% for keyword in metadata['keywords'] -%}
      <div itemprop="keywords" data-type="keyword">{{ keyword }}</div>
      {%- endfor %}
      {% for subject in metadata['subjects'] -%}
      <div itemprop="about" data-type="subject">{{ subject }}</div>
      {%- endfor %}
    </div>

   {{ content }}
  </body>
</html>
"""

# YANK This was pulled from cnx-archive to temporarily provide
#      a way to render the the tree to html. This either needs to
#      move elsewhere or preferably be replaced with a better solution.

def html_listify(tree, root_xl_element, extensions, list_type='ol'):
    for node in tree:
        li_elm = etree.SubElement(root_xl_element, 'li')
        if node['id'] == 'subcol':
            span_elm = etree.SubElement(li_elm, 'span')
            span_elm.text = node['title']
        else:
            a_elm = etree.SubElement(li_elm, 'a')
            a_elm.text = node['title']
            a_elm.set('href', ''.join([node['id'], extensions[node['id']]]))
        if 'contents' in node:
            elm = etree.SubElement(li_elm, list_type)
            html_listify(node['contents'], elm, extensions)


def tree_to_html(tree, extensions):
    nav = etree.Element('nav')
    nav.set('id', 'toc')
    ol = etree.SubElement(nav, 'ol')
    html_listify(tree['contents'], ol, extensions)
    return etree.tostring(nav)

# /YANK



# -*- coding: utf-8 -*-
# ###
# Copyright (c) 2013, Rice University
# This software is subject to the provisions of the GNU Affero General
# Public License version 3 (AGPLv3).
# See LICENCE.txt for details.
# ###
import os
import tempfile
import unittest

from lxml import etree

from .. import testing


here = os.path.abspath(os.path.dirname(__file__))
TEST_DATA_DIR = os.path.join(here, 'data')


class EPUBTestCase(testing.EPUBTestCase):

    @property
    def target_cls(self):
        """Import the target class"""
        from ..epub import EPUB
        return EPUB

    def test_obj_from_unknown_file_type(self):
        """Test that we get an exception when
        an unknown filetype is given.
        """
        # For example purposes, use any file within the test-data directory.
        filepath = os.path.join(TEST_DATA_DIR, 'nav-tree.xhtml')

        with self.assertRaises(TypeError) as caught_exception:
            epub = self.target_cls.from_file(filepath)

    def test_obj_from_directory(self):
        """Test that we can read an (unarchived) epub from the filesystem."""
        epub_filepath = os.path.join(TEST_DATA_DIR, 'blank')

        epub = self.target_cls.from_file(epub_filepath)

        self.assertEqual(epub._root, epub_filepath)

    def test_obj_from_epub_file(self):
        """Test that we can read an .epub file."""
        epub_filepath = self.pack_epub(os.path.join(TEST_DATA_DIR, 'blank'))

        epub = self.target_cls.from_file(epub_filepath)

        # Does it unpack to a temporary location?
        self.assertTrue(epub._root.startswith(tempfile.tempdir))

    def test_obj_from_open_epub_file(self):
        """Test that we can read an open .epub file."""
        epub_filepath = self.pack_epub(os.path.join(TEST_DATA_DIR, 'blank'))

        with open(epub_filepath, 'rb') as zf:
            epub = self.target_cls.from_file(zf)

        # Does it unpack to a temporary location?
        self.assertTrue(epub._root.startswith(tempfile.tempdir))
        from ..epub import (
            EPUB_MIMETYPE_RELATIVE_PATH,
            EPUB_MIMETYPE_CONTENTS,
            )
        mimetype_filepath = os.path.join(epub._root,
                                         EPUB_MIMETYPE_RELATIVE_PATH)
        with open(mimetype_filepath, 'r') as fb:
            contents = fb.read().strip()
            self.assertEqual(contents, EPUB_MIMETYPE_CONTENTS)

    def test_package_parsing(self):
        """Test that packages are parsed into the EPUB.
        This does not examine whether the packages themselves are correct,
        only that the packages are represented within the EPUB object.
        """
        # Use the book test data, which contains a single package with
        # valid values all the way down.
        epub_filepath = self.pack_epub(os.path.join(TEST_DATA_DIR, 'book'))

        # Parse an EPUB.
        epub = self.target_cls.from_file(epub_filepath)

        # EPUBs have packages which the object treats as iterable items.
        self.assertEqual(len(epub), 1)


class PackageTestCase(testing.EPUBTestCase):

    def make_one(self, file):
        from ..epub import Package
        return Package.from_file(file)

    def test_wo_navigation_item(self):
        """Navigation documents are required in a valid EPUB.
        We will enforce this requirement, as we (cnx) also need the navigation
        document in order to build a tree of publishable items.
        """
        # Copy the test data to modify it.
        epub_filepath = os.path.join(TEST_DATA_DIR, 'book')
        epub_root = self.copy(epub_filepath)
        package_filepath = os.path.join(
            epub_root,
            "9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6.opf")
        # Remove the navigation item from the 'book' test data.
        with open(package_filepath, 'r') as fb:
            xml = etree.parse(fb)
        from ..epub import EPUB_OPF_NAMESPACES as opf_nsmap
        manifest = xml.xpath('//opf:manifest', namespaces=opf_nsmap)[0]
        nav_item = xml.xpath('//opf:item[@properties="nav"]',
                             namespaces=opf_nsmap)[0]
        del manifest[manifest.index(nav_item)]
        with open(package_filepath, 'wb') as fb:
            fb.write(etree.tostring(xml))

        from ..epub import MissingNavigationError as Error
        with self.assertRaises(Error) as caught_exception:
            package = self.make_one(package_filepath)

    def test_w_double_navigation_item(self):
        """Navigation documents are required in a valid EPUB.
        We will enfource this requirement, as we (cnx) also need the navigation
        document in order to build a tree of publishable items.
        However, we need to limit the navigation to one per package,
        because in our scheme the package's navigation document defines
        a book tree.
        """
        # Copy the test data to modify it.
        epub_root = self.copy(os.path.join(TEST_DATA_DIR, 'book'))
        package_filepath = os.path.join(
            epub_root,
            "9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6.opf")
        # Remove the navigation item from the 'book' test data.
        with open(package_filepath, 'r') as fb:
            xml = etree.parse(fb)
        from ..epub import EPUB_OPF_NAMESPACES as opf_nsmap
        manifest = xml.xpath('//opf:manifest', namespaces=opf_nsmap)[0]
        nav_item = xml.xpath('//opf:item[@properties="nav"]',
                             namespaces=opf_nsmap)[0]
        from copy import deepcopy
        nav_item = deepcopy(nav_item)
        nav_item.set('id', "toc2")
        manifest.append(nav_item)
        with open(package_filepath, 'wb') as fb:
            fb.write(etree.tostring(xml))

        from ..epub import AdditionalNavigationError as Error
        with self.assertRaises(Error) as caught_exception:
            package = self.make_one(package_filepath)

    def test_metadata_parsing(self):
        """With a completely NEW package
        (NOT based on or derived from any other work),
        check parsed metadata for accuracy.
        """
        package_filepath = os.path.join(
            TEST_DATA_DIR, 'book',
            "9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6.opf")
        package = self.make_one(package_filepath)

        expected_metadata = {
            'publisher': "Connexions",
            'publication_message': u'Nueva Versión',
            'identifier': "org.cnx.contents.9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6",
            'title': "9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6",
            'language': "en-us",
            'publisher': "Connexions",
            'license_text': "This work is shared with the public using the Attribution 3.0 Unported (CC BY 3.0) license.",
            'license_url': "http://creativecommons.org/licenses/by/3.0/",
            }
        self.assertEqual(package.metadata, expected_metadata)

    def test_item_containment(self):
        package_filepath = os.path.join(
            TEST_DATA_DIR, 'book',
            "9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6.opf")
        package = self.make_one(package_filepath)
        # EPUB Packages have contents...
        self.assertEqual(len(package), 3)

        # Check the navigation order
        expected_nav_document = package.grab_by_name(
                "9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6.xhtml")
        self.assertTrue(expected_nav_document.is_navigation)
        self.assertEqual(package.navigation, expected_nav_document)

        # Verify reference to a resource item.
        # This also checks iteration and containment
        resource = [i for i in package
                    if i.name == "e3d625fe893b3f1f9aaef3bdf6bfa15c.png"][0]
        self.assertIn(resource, package)

    def test_parsing_same_file_twice(self):
        # Tests identity issues accross more than one instance.
        epub_filepath = self.pack_epub(os.path.join(TEST_DATA_DIR, 'book'))

        from ..epub import EPUB
        epub1 = EPUB.from_file(epub_filepath)
        epub2 = EPUB.from_file(epub_filepath)

        self.assertEqual(len(epub1), 1)
        self.assertEqual(len(epub1[0]), 3)
        self.assertEqual(len(epub2), 1)
        self.assertEqual(len(epub2[0]), 3)


class WritePackageTestCase(testing.EPUBTestCase):
    """Output the ``Package`` to the filesystem"""

    def test_to_file(self):
        """Write a populated Package to the filesystem."""
        # Use the 'book' data to test against. This enables us to
        # test the resulting structure against the existing structure.

        # Packages are not mutable and shouldn't be, because one wouldn't
        # normally create a package by hand. It would be created via
        # the reading from the filesystem or through adaptation of
        # a ``Binder``'ish object.
        book_path = os.path.join(TEST_DATA_DIR, 'book')
        from ..epub import Item
        items = [
            {'filepath': os.path.join(
                book_path, 'content',
                '9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6.xhtml'),
             'media_type': 'application/xhtml+xml',
             'is_navigation': True,
             'properties': ['nav'],
             },
            {'filepath': os.path.join(
                 book_path, 'content',
                 'e78d4f90-e078-49d2-beac-e95e8be70667@3.xhtml'),
             'media_type': 'application/xhtml+xml',
             },
            {'filepath': os.path.join(
                 book_path, 'resources',
                 'e3d625fe893b3f1f9aaef3bdf6bfa15c.png'),
             'media_type': 'image/png',
             }
            ]
        items = [Item.from_file(**i) for i in items]

        package_name = '9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6.opf'
        package_metadata = {
            'publisher': "Connexions",
            'publication_message': "Loosely publishing these here modules.",
            'title': "9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6",
            'identifier': "org.cnx.contents.9b0903d2-13c4-4ebe-" \
                          "9ffe-1ee79db28482@1.6",
            'language': 'en-us',
            'license_text': "This work is shared with the public using " \
                            "the Attribution 3.0 Unported (CC BY 3.0) " \
                            "license.",
            'license_url': "http://creativecommons.org/licenses/by/3.0/",
            }
        from ..epub import Package
        package = Package(package_name, items, package_metadata)

        output_path = self.tmpdir
        package.to_file(package, output_path)

        # Verify...
        walker = os.walk(output_path)
        for root, dirs, files in walker:
            break
        self.assertEqual(['contents', 'resources'], sorted(dirs))
        self.assertEqual([package_name], files)
        # ./contents/
        files = os.listdir(os.path.join(output_path, 'contents'))
        self.assertEqual(['9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6.xhtml',
                          'e78d4f90-e078-49d2-beac-e95e8be70667@3.xhtml'],
                         sorted(files))
        # ./resources/
        files = os.listdir(os.path.join(output_path, 'resources'))
        self.assertEqual(['e3d625fe893b3f1f9aaef3bdf6bfa15c.png'],
                         sorted(files))

        with open(os.path.join(output_path, package_name), 'r') as fb:
            opf_xml = etree.parse(fb)

        # Parse the file and see if the metadata matches.
        from ..epub import OPFParser
        parser = OPFParser(opf_xml)

        self.assertEqual(parser.metadata, package_metadata)


class WriteEPUBTestCase(testing.EPUBTestCase):
    """Output the ``EPUB`` to the filesystem"""

    def test_to_file(self):
        """Write a populated EPUB to the filesystem as a zipfile."""
        # Use the 'book' data to test against. This enables us to
        # test the resulting structure against the existing structure.

        # EPUB and Package are not mutable and shouldn't be,
        # because one wouldn't normally create an EPUB by hand.
        # It would be created via the reading from the filesystem
        # or through adaptation of a ``Binder``'ish object(s).
        book_path = os.path.join(TEST_DATA_DIR, 'book')
        from ..epub import Item
        items = [
            {'filepath': os.path.join(
                book_path, 'content',
                '9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6.xhtml'),
             'media_type': 'application/xhtml+xml',
             'is_navigation': True,
             'properties': ['nav'],
             },
            {'filepath': os.path.join(
                 book_path, 'content',
                 'e78d4f90-e078-49d2-beac-e95e8be70667@3.xhtml'),
             'media_type': 'application/xhtml+xml',
             },
            {'filepath': os.path.join(
                 book_path, 'resources',
                 'e3d625fe893b3f1f9aaef3bdf6bfa15c.png'),
             'media_type': 'image/png',
             }
            ]
        items = [Item.from_file(**i) for i in items]

        package_name = '9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6.opf'
        package_metadata = {
            'publisher': "Connexions",
            'publication_message': "Loosely publishing these here modules.",
            'title': "9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6",
            'identifier': "org.cnx.contents.9b0903d2-13c4-4ebe-" \
                          "9ffe-1ee79db28482@1.6",
            'language': 'en-us',
            'license_text': "This work is shared with the public using " \
                            "the Attribution 3.0 Unported (CC BY 3.0) " \
                            "license.",
            'license_url': "http://creativecommons.org/licenses/by/3.0/",
            }
        from ..epub import Package
        package = Package(package_name, items, package_metadata)

        from ..epub import EPUB
        epub = EPUB(packages=[package])

        output_path = self.tmpdir
        epub_filename = 'book.epub'
        epub_filepath = os.path.join(output_path, epub_filename)
        epub.to_file(epub, epub_filepath)

        # Unpack so we can check the contents...
        from ..epub import unpack_epub
        unpack_path = os.path.join(self.tmpdir, 'unpacked-epub')
        os.mkdir(unpack_path)
        epub_directory = unpack_epub(epub_filepath, unpack_path)

        # Verify...
        walker = os.walk(unpack_path)
        for root, dirs, files in walker:
            break
        self.assertEqual(['META-INF', 'contents', 'resources'],
                         sorted(dirs))
        self.assertEqual([package_name, 'mimetype'], sorted(files))
        # ./META-INF/
        files = os.listdir(os.path.join(unpack_path, 'META-INF'))
        self.assertEqual(['container.xml'], files)
        # ./contents/
        files = os.listdir(os.path.join(unpack_path, 'contents'))
        self.assertEqual(['9b0903d2-13c4-4ebe-9ffe-1ee79db28482@1.6.xhtml',
                          'e78d4f90-e078-49d2-beac-e95e8be70667@3.xhtml'],
                         sorted(files))
        # ./resources/
        files = os.listdir(os.path.join(unpack_path, 'resources'))
        self.assertEqual(['e3d625fe893b3f1f9aaef3bdf6bfa15c.png'], files)

        from ..epub import EPUB_CONTAINER_XML_RELATIVE_PATH
        container_xml_filepath = os.path.join(unpack_path,
                                              EPUB_CONTAINER_XML_RELATIVE_PATH)
        with open(container_xml_filepath, 'r') as fb:
            container_xml = fb.read()

        with open(os.path.join(unpack_path, 'mimetype'), 'r') as fb:
            self.assertEqual(fb.read(), 'application/epub+zip')

        # Parse the file and check for opf inclusion.
        expected_string = 'full-path="{}"'.format(package_name)
        self.assertTrue(container_xml.find(expected_string) >= 0,
                        container_xml)
